@@ -376,8 +376,7 @@ typedef int sasl_client_plug_init_t(const sasl_utils_t *utils,
 				    int max_version,
 				    int *out_version,
 				    sasl_client_plug_t **pluglist,
-				    int *plugcount,
-				    const char *plugname);
+				    int *plugcount);
 
 /* add a client plug-in
  */
@@ -676,7 +675,7 @@ typedef struct sasl_server_plug {
 
 /* plug-in entry point:
  *  utils         -- utility callback functions
- *  plugname      -- name of plug-in
+ *  plugname      -- name of plug-in (may be NULL)
  *  max_version   -- highest server plug version supported
  * returns:
  *  out_version   -- server plug-in version of result
@@ -693,8 +692,7 @@ typedef int sasl_server_plug_init_t(const sasl_utils_t *utils,
 				    int max_version,
 				    int *out_version,
 				    sasl_server_plug_t **pluglist,
-				    int *plugcount,
-				    const char *plugname);
+				    int *plugcount);
 
 /* 
  * add a server plug-in
@@ -716,7 +714,7 @@ typedef struct sasl_canonuser {
     /* global state for plugin */
     void *glob_context;
 
-    /* spare pointer */
+    /* name of plugin */
     void *spare_ptr1;
 
     /* free global state for plugin */
@@ -737,6 +735,8 @@ typedef struct sasl_canonuser {
      *  out_amax         -- max length of authid
      *  out_alen         -- set to length of authid (0 = same as user name)
      *
+     *  note that the output buffers MAY be the same as the input buffers.
+     *
      * returns
      *  SASL_OK         on success
      *  SASL_BADPROT    username contains invalid character
