@@ -1,6 +1,6 @@
 /* auxprop.c - auxilliary property support
  * Rob Siemborski
- * $Id: auxprop.c,v 1.1.2.11 2001/06/27 14:56:27 rjs3 Exp $
+ * $Id: auxprop.c,v 1.1.2.16 2001/07/12 14:10:11 rjs3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -55,7 +55,7 @@ struct proppool
     size_t unused;        /* Space unused in this pool between end
 			   * of char** area and beginning of char* area */
 
-    char data[0];         /* Variable Sized */
+    char data[1];         /* Variable Sized */
 };
 
 struct propctx  {
@@ -82,7 +82,9 @@ static auxprop_plug_list_t *auxprop_head = NULL;
 static struct proppool *alloc_proppool(size_t size) 
 {
     struct proppool *ret;
-    size_t total_size = sizeof(struct proppool) + size;
+    /* minus 1 for the one that is already a part of the array
+     * in the struct */
+    size_t total_size = sizeof(struct proppool) + size - 1;
     ret = sasl_ALLOC(total_size);
     if(!ret) return NULL;
 
@@ -723,9 +725,12 @@ int prop_setvals(struct propctx *ctx, const char *name,
  */
 int sasl_auxprop_request(sasl_conn_t *conn, const char **propnames) 
 {
+    int result;
     sasl_server_conn_t *sconn;
 
-    if(!conn || conn->type != SASL_CONN_SERVER) return SASL_BADPARAM;
+    if(!conn) return SASL_BADPARAM;
+    if(conn->type != SASL_CONN_SERVER)
+	PARAMERROR(conn);
     
     sconn = (sasl_server_conn_t *)conn;
 
@@ -734,7 +739,8 @@ int sasl_auxprop_request(sasl_conn_t *conn, const char **propnames)
 	return SASL_OK;
     }
     
-    return prop_request(sconn->sparams->propctx, propnames);
+    result = prop_request(sconn->sparams->propctx, propnames);
+    RETURN(conn, result);
 }
 
 
@@ -770,7 +776,7 @@ int sasl_auxprop_add_plugin(const char *plugname,
 			 &out_version, &plug, plugname);
 
     if(result != SASL_OK) {
-	VL(("auxpropfunc error %i\n",result));
+	_sasl_log(NULL, SASL_LOG_ERR, "auxpropfunc error %i\n",result);
 	return result;
     }
 
