@@ -1,7 +1,7 @@
 /* common.c - Functions that are common to server and clinet
  * Rob Siemborski
  * Tim Martin
- * $Id: common.c,v 1.64.2.26 2001/06/27 20:12:10 rjs3 Exp $
+ * $Id: common.c,v 1.64.2.37 2001/07/12 21:04:37 rjs3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -159,11 +159,13 @@ int sasl_encode(sasl_conn_t *conn, const char *input,
 		unsigned inputlen,
 		const char **output, unsigned *outputlen)
 {
+    int result;
     struct iovec tmp;
 
-    if(!conn || !input || !output || !outputlen)
-	return SASL_BADPARAM;
-
+    if(!conn) return SASL_BADPARAM;
+    if(!input || !inputlen || !output || !outputlen)
+	PARAMERROR(conn);
+    
     /* maxoutbuf checking is done in sasl_encodev */
 
     /* Note: We are casting a const pointer here, but it's okay
@@ -172,7 +174,9 @@ int sasl_encode(sasl_conn_t *conn, const char *input,
     tmp.iov_base = (void *)input;
     tmp.iov_len = inputlen;
     
-    return sasl_encodev(conn, &tmp, 1, output, outputlen);
+    result = sasl_encodev(conn, &tmp, 1, output, outputlen);
+
+    RETURN(conn, result);
 }
 
 /* security-encode an iovec */
@@ -185,8 +189,9 @@ int sasl_encodev(sasl_conn_t *conn,
     unsigned i;
     size_t total_size = 0;
 
-    if (! conn || ! invec || ! output || ! outputlen || numiov < 1)
-	return SASL_BADPARAM;
+    if (!conn) return SASL_BADPARAM;
+    if (! invec || ! output || ! outputlen || numiov < 1)
+	PARAMERROR(conn);
 
     /* This might be better to check on a per-plugin basis, but I think
      * it's cleaner and more effective here.  It also encourages plugins
@@ -195,22 +200,22 @@ int sasl_encodev(sasl_conn_t *conn,
     for(i=0; i<numiov;i++) {
 	total_size += invec[i].iov_len;
     }    
-    if(total_size > conn->oparams.maxoutbuf) {
-	return SASL_BADPARAM;
-    }
+    if(total_size > conn->oparams.maxoutbuf)
+	PARAMERROR(conn);
 
     if(conn->oparams.encode == NULL)  {
 	result = _iovec_to_buf(invec, numiov, &conn->encode_buf);
-	if(result != SASL_OK) return result;
+	if(result != SASL_OK) INTERROR(conn, result);
        
 	*output = conn->encode_buf->data;
 	*outputlen = conn->encode_buf->curlen;
 
-        return SASL_OK;
     } else {
-	return conn->oparams.encode(conn->context, invec, numiov,
-				    output, outputlen);
+	result = conn->oparams.encode(conn->context, invec, numiov,
+				      output, outputlen);
     }
+
+    RETURN(conn, result);
 }
  
 /* output is only valid until next call to sasl_decode */
@@ -220,18 +225,23 @@ int sasl_decode(sasl_conn_t *conn,
 {
     int result;
 
-    if(!conn || !input || !output || !outputlen)
-	return SASL_BADPARAM;
+    if(!conn) return SASL_BADPARAM;
+    if(!input || !output || !outputlen)
+	PARAMERROR(conn);
 
     /* FIXME: do we verify maxoutbuf on incoming data as well? */
-    if(inputlen > conn->oparams.maxoutbuf) return SASL_BADPARAM;
-
+    if(inputlen > conn->oparams.maxoutbuf)
+	PARAMERROR(conn);
+    
     if(conn->oparams.decode == NULL)
     {
-	result = _buf_alloc(&conn->decode_buf, &conn->decode_buf_len,
-			    inputlen + 1);
-	if(result != SASL_OK)
-	    return result;
+	/* Since we know how long the output is maximally, we can
+	 * just allocate it to begin with, and never need another
+         * allocation! */
+	if(!conn->decode_buf)
+	    conn->decode_buf = sasl_ALLOC(conn->oparams.maxoutbuf + 1);
+	if(!conn->decode_buf)
+	    MEMERROR( conn );
 	
 	memcpy(conn->decode_buf, input, inputlen);
 	conn->decode_buf[inputlen] = '\0';
@@ -242,10 +252,10 @@ int sasl_decode(sasl_conn_t *conn,
     } else {
         result = conn->oparams.decode(conn->context, input, inputlen,
                                       output, outputlen);
-        return result;
+        RETURN(conn, result);
     }
 
-    return SASL_FAIL;
+    INTERROR(conn, SASL_FAIL);
 }
 
 
@@ -270,7 +280,8 @@ void sasl_done(void)
     _sasl_client_cleanup_hook();
   
   _sasl_canonuser_free();
-
+  _sasl_done_with_plugins();
+  
   sasl_MUTEX_FREE(free_mutex);
   free_mutex = NULL;
 
@@ -296,38 +307,24 @@ int _sasl_conn_init(sasl_conn_t *conn,
 		    const sasl_global_callbacks_t *global_callbacks) {
   int result = SASL_OK;
 
-  I(conn);
-  I(service);
-
   conn->type = SASL_CONN_UNKNOWN;
 
   result = _sasl_strdup(service, &conn->service, NULL);
-  if (result != SASL_OK) return result;
+  if (result != SASL_OK) 
+      MEMERROR(conn);
 
   memset(&conn->oparams, 0, sizeof(sasl_out_params_t));
   memset(&conn->external, 0, sizeof(_sasl_external_properties_t));
 
   conn->secflags = secflags;
 
-  if(!iplocalport ||
-     _sasl_ipfromstring(iplocalport, &conn->ip_local) != SASL_OK) {
-      memset(&conn->ip_local, 0, sizeof(conn->ip_local));
-      conn->got_ip_local = 0;
-  } else {
-      /* It checks out OK */
-      strcpy(conn->iplocalport, iplocalport);
-      conn->got_ip_local = 1;
-  }
+  result = sasl_setprop(conn, SASL_IPLOCALPORT, iplocalport);
+  if(result != SASL_OK)
+      RETURN(conn, result);
   
-  if(!ipremoteport ||
-     _sasl_ipfromstring(ipremoteport, &conn->ip_remote) != SASL_OK) {
-      memset(&conn->ip_remote, 0, sizeof(conn->ip_remote));
-      conn->got_ip_remote = 0;
-  } else {
-      /* It checks out OK */
-      strcpy(conn->ipremoteport, ipremoteport);
-      conn->got_ip_remote = 1;
-  }
+  result = sasl_setprop(conn, SASL_IPREMOTEPORT, ipremoteport);
+  if(result != SASL_OK)
+      RETURN(conn, result);
   
   conn->encode_buf = NULL;
   conn->context = NULL;
@@ -339,13 +336,18 @@ int _sasl_conn_init(sasl_conn_t *conn,
   memset(&conn->props, 0, sizeof(conn->props));
 
   /* Start this buffer out as an empty string */
-  conn->error_buf = NULL;
-  conn->error_buf_len = 0;
-  result = _buf_alloc(&conn->error_buf, &conn->error_buf_len, 100);     
-  if(result != SASL_OK) return result;
+  conn->error_code = SASL_OK;
+  conn->errdetail_buf = conn->error_buf = NULL;
+  conn->errdetail_buf_len = conn->error_buf_len = 150;
 
+  result = _buf_alloc(&conn->error_buf, &conn->error_buf_len, 150);     
+  if(result != SASL_OK) MEMERROR(conn);
+  result = _buf_alloc(&conn->errdetail_buf, &conn->errdetail_buf_len, 150);
+  if(result != SASL_OK) MEMERROR(conn);
+  
   conn->error_buf[0] = '\0';
-
+  conn->errdetail_buf[0] = '\0';
+  
   conn->decode_buf = NULL;
   conn->decode_buf_len = 0;
 
@@ -360,7 +362,9 @@ int _sasl_conn_init(sasl_conn_t *conn,
     result = _sasl_strdup(serverFQDN, &conn->serverFQDN, NULL);
   }
 
-  return result;
+  if(result != SASL_OK) MEMERROR( conn );
+
+  RETURN(conn, SASL_OK);
 }
 
 /* It turns out to be conveinent to have a shared sasl_utils_t */
@@ -427,6 +431,9 @@ void _sasl_conn_dispose(sasl_conn_t *conn) {
   if(conn->error_buf)
       sasl_FREE(conn->error_buf);
   
+  if(conn->errdetail_buf)
+      sasl_FREE(conn->errdetail_buf);
+
   if(conn->decode_buf)
       sasl_FREE(conn->decode_buf);
 
@@ -452,8 +459,8 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
   sasl_getopt_t *getopt;
   void *context;
   
-  if (! conn) return SASL_FAIL;
-  if (! pvalue) return SASL_FAIL;
+  if (! conn) return SASL_BADPARAM;
+  if (! pvalue) PARAMERROR(conn);
 
   switch(propnum)
   {
@@ -465,7 +472,7 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
       break;
   case SASL_GETOPTCTX:
       result = _sasl_getcallback(conn, SASL_CB_GETOPT, &getopt, &context);
-      if(result != SASL_OK) return result;
+      if(result != SASL_OK) break;
       
       *(void **)pvalue = context;
       break;
@@ -473,16 +480,20 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
       *(const sasl_callback_t **)pvalue = conn->callbacks;
       break;
   case SASL_IPLOCALPORT:
-      if (! conn->got_ip_local)
-	result = SASL_NOTDONE;
-      else
-	*(const char **)pvalue = conn->iplocalport;
+      if(conn->got_ip_local)
+	  *(const char **)pvalue = conn->iplocalport;
+      else {
+	  *(const char **)pvalue = NULL;
+	  result = SASL_NOTDONE;
+      }
       break;
   case SASL_IPREMOTEPORT:
-      if (! conn->got_ip_remote)
-	result = SASL_NOTDONE;
-      else
-	*(const char **)pvalue = conn->ipremoteport;
+      if(conn->got_ip_remote)
+	  *(const char **)pvalue = conn->ipremoteport;
+      else {
+	  *(const char **)pvalue = NULL;
+	  result = SASL_NOTDONE;
+      }	  
       break;
   case SASL_USERNAME:
       if(! conn->oparams.user)
@@ -501,6 +512,7 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
   case SASL_SERVICE:
       *((const char **)pvalue) = conn->service;
       break;
+  case SASL_AUTHSOURCE: /* FIXME: I doubt these are identical */
   case SASL_MECHNAME:
       if(conn->type == SASL_CONN_CLIENT) {
 	  *((const char **)pvalue) = ((sasl_client_conn_t *)conn)->mech->plug->mech_name;
@@ -512,10 +524,6 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
       
       if(!(*pvalue) && result == SASL_OK) result = SASL_NOTDONE;
       break;
-  case SASL_AUTHSOURCE:
-      /* FIXME: What is this? */
-      result=SASL_FAIL;
-      break;
   case SASL_PLUGERR:
       *((const char **)pvalue) = conn->error_buf;
       break;
@@ -532,7 +540,12 @@ int sasl_getprop(sasl_conn_t *conn, int propnum, const void **pvalue)
       result = SASL_BADPARAM;
   }
 
-  return result; 
+  if(result == SASL_BADPARAM) {
+      PARAMERROR(conn);
+  } else if(result != SASL_OK) {
+      INTERROR(conn, result);
+  } else
+      RETURN(conn, result); 
 }
 
 /* set property in SASL connection state
@@ -554,10 +567,11 @@ int sasl_setprop(sasl_conn_t *conn, int propnum, const void *value)
   case SASL_SSF_EXTERNAL:
       conn->external.ssf = *((sasl_ssf_t *)value);
       break;
+
   case SASL_AUTH_EXTERNAL:
       if(value && strlen(value)) {
 	  result = _sasl_strdup(value, &str, NULL);
-	  if(result != SASL_OK) return result;
+	  if(result != SASL_OK) MEMERROR(conn);
       } else {
 	  str = NULL;
       }
@@ -568,15 +582,50 @@ int sasl_setprop(sasl_conn_t *conn, int propnum, const void *value)
       conn->external.auth_id = str;
 
       break;
+
   case SASL_SEC_PROPS:
       memcpy(&(conn->props),(sasl_security_properties_t *)value,
 	     sizeof(sasl_security_properties_t));
       break;
+
+  case SASL_IPREMOTEPORT:
+  {
+      const char *ipremoteport = (const char *)value;
+      if(!value) {
+	  conn->got_ip_remote = 0; 
+      } else if (_sasl_ipfromstring(ipremoteport, NULL)
+		 != SASL_OK) {
+	  sasl_seterror(conn, 0, "Bad IPREMOTEPORT value");
+	  RETURN(conn, SASL_BADPARAM);
+      } else {
+	  strcpy(conn->ipremoteport, ipremoteport);
+	  conn->got_ip_remote = 1;
+      }
+      break;
+  }
+
+  case SASL_IPLOCALPORT:
+  {
+      const char *iplocalport = (const char *)value;
+      if(!value) {
+	  conn->got_ip_local = 0;	  
+      } else if (_sasl_ipfromstring(iplocalport, NULL)
+		 != SASL_OK) {
+	  sasl_seterror(conn, 0, "Bad IPLOCALPORT value");
+	  RETURN(conn, SASL_BADPARAM);
+      } else {
+	  strcpy(conn->iplocalport, iplocalport);
+	  conn->got_ip_local = 1;
+      }
+      break;
+  }
+
   default:
+      sasl_seterror(conn, 0, "Unknown parameter type");
       result = SASL_BADPARAM;
   }
   
-  return result;
+  RETURN(conn, result);
 }
 
 /* this is apparently no longer a user function */
@@ -640,9 +689,22 @@ const char *sasl_errstring(int saslerr,
  * a connection */
 const char *sasl_errdetail(sasl_conn_t *conn) 
 {
-    /* FIXME: This should be different from getprop(SASL_PLUGERR) in
-     * that it takes the return code into effect */
-    return conn->error_buf;
+    unsigned need_len;
+    const char *errstr;
+    char leader[128];
+    
+    errstr = sasl_errstring(conn->error_code, NULL, NULL);
+    snprintf(leader,128,"SASL(%d): %s: ",
+	     sasl_usererr(conn->error_code), errstr);
+    
+    need_len = strlen(leader) + strlen(conn->error_buf) + 12;
+    _buf_alloc(&conn->errdetail_buf, &conn->errdetail_buf_len, need_len);
+
+    snprintf(conn->errdetail_buf, need_len, "%s%s", leader, conn->error_buf);
+   
+	printf("I HAVE %s\n", conn->errdetail_buf);
+
+    return conn->errdetail_buf;
 }
 
 
@@ -1002,7 +1064,8 @@ _sasl_proxy_policy(sasl_conn_t *conn,
 		   unsigned urlen __attribute__((unused)),
 		   struct propctx *propctx __attribute__((unused)))
 {
-    if (!conn) return SASL_BADPARAM;
+    if (!conn)
+	return SASL_BADPARAM;
 
     if (!requested_user || *requested_user == '\0')
 	return SASL_OK;
@@ -1011,28 +1074,29 @@ _sasl_proxy_policy(sasl_conn_t *conn,
 	(memcmp(auth_identity, requested_user, rlen) != 0)) {
 	sasl_seterror(conn, 0,
 		      "Requested identity not authenticated identity");
-	return SASL_BADAUTH;
+	RETURN(conn, SASL_BADAUTH);
     }
 
     return SASL_OK;
 }
 
-int
-_sasl_getcallback(sasl_conn_t * conn,
-		  unsigned long callbackid,
-		  int (**pproc)(),
-		  void **pcontext)
+int _sasl_getcallback(sasl_conn_t * conn,
+		      unsigned long callbackid,
+		      int (**pproc)(),
+		      void **pcontext)
 {
   const sasl_callback_t *callback;
 
+  if (!conn) return SASL_BADPARAM;
   if (! pproc || ! pcontext)
-    return SASL_BADPARAM;
+    PARAMERROR(conn);
 
   /* Some callbacks are always provided by the library */
   switch (callbackid) {
   case SASL_CB_LIST_END:
     /* Nothing ever gets to provide this */
-    return SASL_FAIL;
+      sasl_seterror(conn, 0, "Someone attempted lookup of SASL_CB_LIST_END");
+      RETURN(conn, SASL_FAIL);
   case SASL_CB_GETOPT:
       if (conn) {
 	  *pproc = &_sasl_conn_getopt;
@@ -1108,7 +1172,8 @@ _sasl_getcallback(sasl_conn_t * conn,
   /* Unable to find a callback... */
   *pproc = NULL;
   *pcontext = NULL;
-  return SASL_FAIL;
+  sasl_seterror(conn, SASL_NOLOG, "Unable to find a callback: %d", callbackid);
+  RETURN(conn,SASL_FAIL);
 }
 
 
@@ -1127,7 +1192,7 @@ _sasl_log (sasl_conn_t *conn,
 	   const char *fmt,
 	   ...)
 {
-  char *out=(char *) sasl_ALLOC(100);
+  char *out=(char *) sasl_ALLOC(250);
   int alloclen=100; /* current allocated length */
   int outlen=0; /* current length of output buffer */
   int pos=0; /* current position in format string */
@@ -1257,7 +1322,6 @@ _sasl_log (sasl_conn_t *conn,
     }
   }
 
-
   out[outlen]=0; /* put 0 at end */
 
   va_end(ap);    
@@ -1489,7 +1553,7 @@ int _iovec_to_buf(const struct iovec *vec,
 
     if(ret != SASL_OK) return SASL_NOMEM;
     
-    bzero(out->data, out->reallen);
+    memset(out->data, 0, out->reallen);
     pos = out->data;
     
     for(i=0; i<numiov; i++) {
@@ -1511,7 +1575,7 @@ int _sasl_iptostring(const struct sockaddr_in *addr,
     if(outlen <= 21) return SASL_BUFOVER;
     if(!addr || !out) return SASL_BADPARAM;
 
-    bzero(out,outlen);
+    memset(out, 0, outlen);
 
     for(i=3; i>=0; i--) {
 	a[i] = (addr->sin_addr.s_addr & (0xFF << (8*i))) >> (i*8);
@@ -1519,7 +1583,7 @@ int _sasl_iptostring(const struct sockaddr_in *addr,
     
     snprintf(out,outlen,"%d.%d.%d.%d;%d",(int)a[3],(int)a[2],
 	                                 (int)a[1],(int)a[0],
-	                                 (int)addr->sin_port);
+	                                 (int)ntohs(addr->sin_port));
 
     return SASL_OK;
 }
@@ -1531,7 +1595,9 @@ int _sasl_ipfromstring(const char *addr, struct sockaddr_in *out)
     unsigned int val = 0;
     unsigned int port;
     
-    if(!addr || !out) return SASL_BADPARAM;
+    /* A NULL out pointer just implies we don't do a copy, just verify it */
+
+    if(!addr) return SASL_BADPARAM;
 
     /* Parse the address */
     for(i=0; i<4 && *addr && *addr != ';'; i++) {
@@ -1544,7 +1610,7 @@ int _sasl_ipfromstring(const char *addr, struct sockaddr_in *out)
 	val |= inval;
 	
         for(;*addr && *addr != '.' && *addr != ';'; addr++)
-	    if(!isdigit(*addr)) return SASL_BADPARAM;
+	    if(!isdigit((int)(*addr))) return SASL_BADPARAM;
 
 	/* skip the separator */
 	addr++;
@@ -1560,11 +1626,13 @@ int _sasl_ipfromstring(const char *addr, struct sockaddr_in *out)
     if((port & 0xFFFF) != port) return SASL_BADPARAM;
         
     for(;*addr;addr++)
-	if(!isdigit(*addr)) return SASL_BADPARAM;
-    
-    bzero(out, sizeof(struct sockaddr_in));
-    out->sin_addr.s_addr = val;
-    out->sin_port = port;
+	if(!isdigit((int)(*addr))) return SASL_BADPARAM;
 
+    if(out) {
+	memset(out, 0, sizeof(struct sockaddr_in));
+	out->sin_addr.s_addr = val;
+	out->sin_port = htons(port);
+    }
+    
     return SASL_OK;
 }
