@@ -1,7 +1,7 @@
 /* SASL server API implementation
  * Rob Siemborski
  * Tim Martin
- * $Id: server.c,v 1.84.2.32 2001/06/27 15:55:44 rjs3 Exp $
+ * $Id: server.c,v 1.84.2.44 2001/07/12 14:10:11 rjs3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -87,7 +87,7 @@ extern int gethostname(char *, int);
  * sasl_server_step
  * sasl_checkpass
  * sasl_checkapop
- * sasl_user_exists <= not yet implemented
+ * sasl_user_exists
  * sasl_setpass
  */
 
@@ -109,14 +109,18 @@ static sasl_global_callbacks_t global_callbacks;
  *  user        -- user name
  *  pass        -- plaintext password, may be NULL to remove user
  *  passlen     -- length of password, 0 = strlen(pass)
+ *  oldpass     -- NULL will sometimes work
+ *  oldpasslen  -- length of password, 0 = strlen(oldpass)
  *  flags       -- see flags below
- *  errstr      -- optional more detailed error
  * 
  * returns:
  *  SASL_NOCHANGE  -- proper entry already exists
  *  SASL_NOMECH    -- no authdb supports password setting as configured
+ *  SASL_NOVERIFY  -- user exists, but no settable password present
  *  SASL_DISABLED  -- account disabled
  *  SASL_PWLOCK    -- password locked
+ *  SASL_WEAKPASS  -- password too weak for security policy
+ *  SASL_NOUSERPASS -- user-supplied passwords not permitted
  *  SASL_FAIL      -- OS error
  *  SASL_BADPARAM  -- password too long
  *  SASL_OK        -- successful
@@ -125,26 +129,26 @@ static sasl_global_callbacks_t global_callbacks;
 int sasl_setpass(sasl_conn_t *conn,
 		 const char *user,
 		 const char *pass, unsigned passlen,
-		 const char *oldpass __attribute__((unused)),
-		 unsigned oldpasslen __attribute__((unused)),
+		 const char *oldpass,
+		 unsigned oldpasslen,
 		 unsigned flags)
 {
     int result=SASL_OK, tmpresult;
     sasl_server_conn_t *s_conn = (sasl_server_conn_t *) conn;
+    sasl_server_userdb_setpass_t *setpass_cb = NULL;
+    void *context = NULL;
     mechanism_t *m;
      
-    if (!mechlist) return SASL_NOTINIT;
+    if (!_sasl_server_active || !mechlist) return SASL_NOTINIT;
 
     /* check params */
     if (!conn) return SASL_BADPARAM;
      
-    if (!(flags & SASL_SET_DISABLE) && passlen == 0)
-	return SASL_BADPARAM;
+    if ((!(flags & SASL_SET_DISABLE) && passlen == 0)
+        || ((flags & SASL_SET_CREATE) && (flags & SASL_SET_DISABLE)))
+	PARAMERROR(conn);
 
-    if ((flags & SASL_SET_CREATE) && (flags & SASL_SET_DISABLE))
-	return SASL_BADPARAM;
-
-    /* set/create password for PLAIN usage */
+    /* set/create password for sasldb */
     tmpresult = _sasl_sasldb_set_pass(conn, user, pass, passlen, 
 				      s_conn->user_realm, flags);
 
@@ -163,6 +167,24 @@ int sasl_setpass(sasl_conn_t *conn,
 		  "set secret for %s", user);
     }
 
+    /* call userdb callback function */
+    result = _sasl_getcallback(conn, SASL_CB_SERVER_USERDB_SETPASS,
+			       &setpass_cb, &context);
+    if(result == SASL_OK && setpass_cb) {
+	tmpresult = setpass_cb(conn, context, user, pass, passlen,
+			    s_conn->sparams->propctx, flags);
+	if(tmpresult != SASL_OK) {
+	    _sasl_log(conn, SASL_LOG_ERR,
+		      "setpass callback failed for %s: %z",
+		      user, tmpresult);
+	} else {
+	    _sasl_log(conn, SASL_LOG_NOTE,
+		      "setpass callback succeeded for %s", user);
+	}
+    } else {
+	result = SASL_OK;
+    }
+
     /* copy info into sparams */
     s_conn->sparams->serverFQDN = conn->serverFQDN;
     s_conn->sparams->service = conn->service;
@@ -179,7 +201,7 @@ int sasl_setpass(sasl_conn_t *conn,
 				     user,
 				     pass,
 				     passlen,
-				     NULL, 0,
+				     oldpass, oldpasslen,
 				     flags);
 	if (tmpresult == SASL_OK) {
 	    _sasl_log(conn, SASL_LOG_NOTE,
@@ -206,7 +228,7 @@ int sasl_setpass(sasl_conn_t *conn,
 	}
     }
 
-    return result;
+    RETURN(conn, result);
 }
 
 /* local mechanism which disposes of server */
@@ -263,9 +285,9 @@ static int init_mechlist(void)
 /*
  * parameters:
  *  p - entry point
- *  library - shared library ptr returned by dlopen
  */
-static int add_plugin(void *p, void *library) 
+int sasl_server_add_plugin(const char *plugname __attribute__((unused)),
+			   sasl_server_plug_init_t *p)
 {
     int plugcount;
     sasl_server_plug_t *pluglist;
@@ -282,10 +304,11 @@ static int add_plugin(void *p, void *library)
     /* call into the shared library asking for information about it */
     /* version is filled in with the version of the plugin */
     result = entry_point(mechlist->utils, SASL_SERVER_PLUG_VERSION, &version,
-			 &pluglist, &plugcount, NULL);
+			 &pluglist, &plugcount);
 
     if ((result != SASL_OK) && (result != SASL_NOUSER)) {
-	VL(("entry_point error %i\n",result));
+	_sasl_log(NULL, SASL_LOG_WARN,
+		  "server add_plugin entry_point error %i\n", result);
 	return result;
     }
 
@@ -308,17 +331,6 @@ static int add_plugin(void *p, void *library)
 	/* wheather this mech actually has any users in it's db */
 	mech->condition = result; /* SASL_OK or SASL_NOUSER */
 
-	/*
-	 * We want plugin library to close but we only
-	 * want to do this once per plugin regardless
-	 * of how many mechs are in a single plugin 
-	 */
-	if (lupe==0) {
-	    mech->u.library=library;
- 	} else {
-	    mech->u.library=NULL;
-	}
-
 	mech->next = mechlist->mech_list;
 	mechlist->mech_list = mech;
 	mechlist->mech_length++;
@@ -327,12 +339,6 @@ static int add_plugin(void *p, void *library)
     return SASL_OK;
 }
 
-int sasl_server_add_plugin(const char *plugname __attribute__((unused)),
-			   sasl_server_plug_init_t *entry)
-{
-    return add_plugin(entry, NULL);
-}
-
 static void server_done(void) {
   mechanism_t *m;
   mechanism_t *prevm;
@@ -354,11 +360,7 @@ static void server_done(void) {
 	  if (prevm->plug->glob_context!=NULL) {
 	      sasl_FREE(prevm->plug->glob_context);
 	  }
-	  
-	  if (prevm->condition == SASL_OK && prevm->u.library != NULL) {
-	      _sasl_done_with_plugin(prevm->u.library);
-	  }
-	  
+	  	  
 	  sasl_FREE(prevm);    
       }
       _sasl_free_utils(&mechlist->utils);
@@ -535,7 +537,7 @@ static int parse_mechlist_file(const char *mechlistfile)
 	*/
 	
 	/* grab file */
-	n->u.f = grab_field(buf, &ptr);
+	n->f = grab_field(buf, &ptr);
 
 	/* grab mech_name */
 	nplug->mech_name = grab_field(ptr, &ptr);
@@ -578,7 +580,8 @@ static int parse_mechlist_file(const char *mechlistfile)
 }
 
 /* initialize server drivers, done once per process
- *  callbacks      -- base callbacks for all server connections
+ *  callbacks      -- callbacks for all server connections; must include
+ *                    getopt callback
  *  appname        -- name of calling application (for lower level logging)
  * results:
  *  state          -- server state
@@ -598,6 +601,13 @@ int sasl_server_init(const sasl_callback_t *callbacks,
     sasl_getopt_t *getopt;
     void *context;
 
+    const add_plugin_list_t ep_list[] = {
+	{ "sasl_server_plug_init", (void *)&sasl_server_add_plugin },
+	{ "sasl_auxprop_plug_init", (void *)&sasl_auxprop_add_plugin },
+	{ "sasl_canonuser_init", (void *)&sasl_canonuser_add_plugin },
+	{ NULL, NULL }
+    };
+
     /* we require the appname to be non-null */
     if (appname==NULL) return SASL_BADPARAM;
 
@@ -628,19 +638,22 @@ int sasl_server_init(const sasl_callback_t *callbacks,
     /* check db */
     ret = _sasl_server_check_db(vf);
 
-    /* load plugins */
+    /* load internal plugins */
     sasl_auxprop_add_plugin("SASLDB", &sasldb_auxprop_plug_init);
-    add_plugin((void *)&external_server_init, NULL);
+    sasl_server_add_plugin(NULL, &external_server_init);
 
-    /* delayed loading of plugins? */
-    if (_sasl_getcallback(NULL, SASL_CB_GETOPT, &getopt, &context) 
+    /* delayed loading of plugins? (DSO only, as it doesn't
+     * make much [any] sense to delay in the static library case) */
+    if (!_is_sasl_server_static &&
+	_sasl_getcallback(NULL, SASL_CB_GETOPT, &getopt, &context) 
 	   == SASL_OK) {
 	getopt(context, NULL, "plugin_list", &pluginfile, NULL);
     }
+    
     if (pluginfile != NULL) {
 	/* this file should contain a list of plugins available.
 	   we'll load on demand. */
-
+	
 	/* Ask the application if it's safe to use this file */
 	ret = ((sasl_verifyfile_t *)(vf->proc))(vf->context,
 						pluginfile,
@@ -649,16 +662,15 @@ int sasl_server_init(const sasl_callback_t *callbacks,
 	    _sasl_log(NULL, SASL_LOG_ERR,
 		      "unable to load plugin list %s: %z", pluginfile, ret);
 	}
-
+	
 	if (ret == SASL_OK) {
 	    ret = parse_mechlist_file(pluginfile);
 	}
     } else {
 	/* load all plugins now */
-	ret = _sasl_get_mech_list("sasl_server_plug_init",
-				  _sasl_find_getpath_callback(callbacks),
-				  _sasl_find_verifyfile_callback(callbacks),
-				  &add_plugin);
+	ret = _sasl_load_plugins(ep_list,
+				 _sasl_find_getpath_callback(callbacks),
+				 _sasl_find_verifyfile_callback(callbacks));
     }
 
     if (ret == SASL_OK)	ret = _sasl_common_init();
@@ -688,13 +700,14 @@ _sasl_transition(sasl_conn_t * conn,
 {
     const char *dotrans = "n";
     sasl_getopt_t *getopt;
+    int result = SASL_OK;
     void *context;
 
     if (! conn)
 	return SASL_BADPARAM;
 
     if (! conn->oparams.authid)
-	return SASL_NOTDONE;
+	PARAMERROR(conn);
 
     /* check if this is enabled: default to false */
     if (_sasl_getcallback(conn, SASL_CB_GETOPT, &getopt, &context) == SASL_OK)
@@ -706,24 +719,29 @@ _sasl_transition(sasl_conn_t * conn,
     if (*dotrans == '1' || *dotrans == 'y' ||
 	(*dotrans == 'o' && dotrans[1] == 'n') || *dotrans == 't') {
 	/* ok, it's on! */
-	return sasl_setpass(conn,
-			    conn->oparams.authid,
-			    pass,
-			    passlen,
-			    NULL, 0, 0);
+	result = sasl_setpass(conn,
+			      conn->oparams.authid,
+			      pass,
+			      passlen,
+			      NULL, 0, 0);
     }
 
-    return SASL_OK;
+    RETURN(conn,result);
 }
 
 
 /* create context for a single SASL connection
  *  service        -- registered name of the service using SASL (e.g. "imap")
- *  serverFQDN     -- Fully qualified server domain name.  NULL means use
- *                    gethostname().  Useful for multi-homed servers.
- *  user_realm     -- permits multiple user domains on server, NULL = default
+ *  serverFQDN     -- Fully qualified domain name of server.  NULL means use
+ *                    gethostname() or equivalent.
+ *                    Useful for multi-homed servers.
+ *  user_realm     -- permits multiple user realms on server, NULL = default
+ *  iplocalport    -- server IPv4/IPv6 domain literal string with port
+ *                    (if NULL, then mechanisms requiring IPaddr are disabled)
+ *  ipremoteport   -- client IPv4/IPv6 domain literal string with port
+ *                    (if NULL, then mechanisms requiring IPaddr are disabled)
  *  callbacks      -- callbacks (e.g., authorization, lang, new getopt context)
- *  secflags       -- security flags (see above)
+ *  flags          -- usage flags (see above)
  * returns:
  *  pconn          -- new connection context
  *
@@ -756,7 +774,9 @@ int sasl_server_new(const char *service,
 			   &server_idle, serverFQDN,
 			   iplocalport, ipremoteport,
 			   callbacks, &global_callbacks);
-  if (result != SASL_OK) return result;
+  if (result != SASL_OK) {
+      goto done_error;
+  }
 
   serverconn = (sasl_server_conn_t *)*pconn;
 
@@ -769,19 +789,27 @@ int sasl_server_new(const char *service,
 
   /* make sparams */
   serverconn->sparams=sasl_ALLOC(sizeof(sasl_server_params_t));
-  if (serverconn->sparams==NULL) return SASL_NOMEM;
+  if (serverconn->sparams==NULL)
+      MEMERROR(*pconn);
+
   memset(serverconn->sparams, 0, sizeof(sasl_server_params_t));
 
   /* set util functions - need to do rest */
   utils=_sasl_alloc_utils(*pconn, &global_callbacks);
-  if (!utils)
-    return SASL_NOMEM;
+  if (!utils) {
+      result = SASL_NOMEM;
+      goto done_error;
+  }
+  
   utils->checkpass = &sasl_checkpass;
 
   /* Setup the propctx -> We'll assume the default size */
   serverconn->sparams->propctx=prop_new(0);
-  if(!serverconn->sparams->propctx) return SASL_NOMEM;
-
+  if(!serverconn->sparams->propctx) {
+      result = SASL_NOMEM;
+      goto done_error;
+  }
+  
   serverconn->sparams->utils = utils;
   serverconn->sparams->transition = &_sasl_transition;
   serverconn->sparams->canon_user = &_sasl_canon_user;
@@ -810,6 +838,7 @@ int sasl_server_new(const char *service,
 
   if(result == SASL_OK) return SASL_OK;
 
+ done_error:
   _sasl_conn_dispose(*pconn);
   sasl_FREE(*pconn);
   *pconn = NULL;
@@ -830,15 +859,22 @@ static int mech_permitted(sasl_conn_t *conn,
     int myflags;
     sasl_ssf_t minssf = 0;
 
+    if(!conn) return SASL_BADPARAM;
+    
     /* Can this plugin meet the application's security requirements? */
-    if (! plug || ! conn)
+    if (! plug ) {
+	PARAMERROR(conn);
 	return 0;
-
+    }
+    
     if (plug == &external_server_mech) {
 	/* Special case for the external mechanism */
 	if (conn->props.min_ssf > conn->external.ssf
-	    || ! conn->external.auth_id)
+	    || ! conn->external.auth_id) {
+	    sasl_seterror(conn, SASL_NOLOG,
+			  "External SSF not good enough");
 	    return 0;
+	}
     } else {
 	if (conn->props.min_ssf < conn->external.ssf) {
 	    minssf = 0;
@@ -847,23 +883,38 @@ static int mech_permitted(sasl_conn_t *conn,
 	}
 
 	/* Generic mechanism */
-	if (plug->max_ssf < minssf)
+	if (plug->max_ssf < minssf) {
+	    sasl_seterror(conn, SASL_NOLOG,
+			  "mech %s is too weak", plug->mech_name);
 	    return 0; /* too weak */
+	}
+	
     }
 
     /* FIXME: it's probabally not valid to call this more than once per
      * connection context */
     if(plug->mech_avail
        && plug->mech_avail(plug->glob_context,
-			   s_conn->sparams, (void **)&conn) != SASL_OK )
+			   s_conn->sparams, (void **)&conn) != SASL_OK ) {
+	sasl_seterror(conn, SASL_NOLOG,
+		      "mech %s not available",
+		      plug->mech_name);
 	return 0;
+    }
 
     /* Generic mechanism */
-    if (plug->max_ssf < minssf) return 0; /* too weak */
+    if (plug->max_ssf < minssf) {
+	sasl_seterror(conn, SASL_NOLOG,
+		      "too weak");
+	return 0; /* too weak */
+    }
 
     /* if there are no users in the secrets database we can't use this 
        mechanism */
-    if (mech->condition == SASL_NOUSER) return 0;
+    if (mech->condition == SASL_NOUSER) {
+	sasl_seterror(conn, 0, "no users in secrets db");
+	return 0;
+    }
     
     /* security properties---if there are any flags that differ and are
        in what the connection are requesting, then fail */
@@ -879,12 +930,17 @@ static int mech_permitted(sasl_conn_t *conn,
 
     /* do we want to special case SASL_SEC_PASS_CREDENTIALS? nah.. */
     if (((myflags ^ plug->security_flags) & myflags) != 0) {
+	sasl_seterror(conn, 0,
+		      "security flags do not match required");
 	return 0;
     }
 
     /* Check Features */
     if(plug->features & SASL_FEAT_GETSECRET) {
 	/* We no longer support sasl_server_{get,put}secret */
+	sasl_seterror(conn, 0,
+		      "mech %s requires unprovided secret facility",
+		      plug->mech_name);
 	return 0;
     }
 
@@ -907,7 +963,7 @@ static int do_authorization(sasl_server_conn_t *s_conn)
     /* check the proxy callback */
     if (_sasl_getcallback(&s_conn->base, SASL_CB_PROXY_POLICY,
 			  &authproc, &auth_context) != SASL_OK) {
-	return SASL_NOAUTHZ;
+	INTERROR(&s_conn->base, SASL_NOAUTHZ);
     }
 
     ret = authproc(&(s_conn->base), auth_context,
@@ -917,21 +973,22 @@ static int do_authorization(sasl_server_conn_t *s_conn)
 		   (s_conn->user_realm ? strlen(s_conn->user_realm) : 0),
 		   s_conn->sparams->propctx);
 
-    return ret;
+    RETURN(&s_conn->base, ret);
 }
 
 
 /* start a mechanism exchange within a connection context
  *  mech           -- the mechanism name client requested
- *  clientin       -- client initial response, NULL if empty
+ *  clientin       -- client initial response (NUL terminated), NULL if empty
  *  clientinlen    -- length of initial response
- *  serverout      -- initial server challenge, NULL if done
+ *  serverout      -- initial server challenge, NULL if done 
+ *                    (library handles freeing this string)
  *  serveroutlen   -- length of initial server challenge
  * output:
  *  pconn          -- the connection negotiation state on success
- *  errstr         -- set to string to send to user on failure
  *
- * Same returns as sasl_server_step()
+ * Same returns as sasl_server_step() or
+ * SASL_NOMECH if mechanism not available.
  */
 int sasl_server_start(sasl_conn_t *conn,
 		      const char *mech,
@@ -949,9 +1006,10 @@ int sasl_server_start(sasl_conn_t *conn,
     m=mechlist->mech_list;
 
     /* check parameters */
-    if ((conn == NULL) || (mech==NULL)
-	|| ((clientin==NULL) && (clientinlen>0)))
-	return SASL_BADPARAM;
+    if(!conn) return SASL_BADPARAM;
+    
+    if (!mech || ((clientin==NULL) && (clientinlen>0)))
+	PARAMERROR(conn);
 
     while (m!=NULL)
     {
@@ -963,6 +1021,7 @@ int sasl_server_start(sasl_conn_t *conn,
     }
   
     if (m==NULL) {
+	sasl_seterror(conn, 0, "Couldn't find mech %s", mech);
 	result = SASL_NOMECH;
 	goto done;
     }
@@ -981,13 +1040,20 @@ int sasl_server_start(sasl_conn_t *conn,
 	int l = 0;
 
 	/* need to load this plugin */
-	result = _sasl_get_plugin(m->u.f, "sasl_server_plug_init",
+	result = _sasl_get_plugin(m->f,
 		    _sasl_find_verifyfile_callback(global_callbacks.callbacks),
-				  (void **) &entry_point, &library);
+				  &library);
+
+	if (result == SASL_OK) {
+	    result = _sasl_locate_entry(library, "sasl_server_plug_init",
+					(void **)&entry_point);
+	}
+
 	if (result == SASL_OK) {
 	    result = entry_point(mechlist->utils, SASL_SERVER_PLUG_VERSION,
-				 &version, &pluglist, &plugcount, NULL);
+				 &version, &pluglist, &plugcount);
 	}
+
 	if (result == SASL_OK) {
 	    /* find the correct mechanism in this plugin */
 	    for (l = 0; l < plugcount; l++) {
@@ -1013,15 +1079,11 @@ int sasl_server_start(sasl_conn_t *conn,
 	    sasl_FREE((sasl_server_plug_t *) m->plug);
 	    m->plug = &pluglist[l];
 	    m->condition = SASL_OK;
-	    m->u.library = library;
 	}
 
 	if (result != SASL_OK) {
-	    if (library) {
-		/* won't be using you after all */
-		_sasl_done_with_plugin(library);
-	    }
-	    return result;
+	    /* The library will eventually be freed, don't sweat it */
+	    RETURN(conn, result);
 	}
     }
 
@@ -1048,7 +1110,7 @@ int sasl_server_start(sasl_conn_t *conn,
     }
 
  done:
-    return result;
+    RETURN(conn,result);
 }
 
 
@@ -1057,8 +1119,7 @@ int sasl_server_start(sasl_conn_t *conn,
  *                      NULL on first step if no optional client step
  *  outputlen & output -- set to the server data to transmit
  *                        to the client in the next step
- *  errstr           -- set to a more text error message from
- *                    a lower level mechanism on failure
+ *                        (library handles freeing this)
  *
  * returns:
  *  SASL_OK        -- exchange is complete.
@@ -1080,8 +1141,9 @@ int sasl_server_step(sasl_conn_t *conn,
     sasl_server_conn_t *s_conn = (sasl_server_conn_t *) conn;  /* cast */
 
     /* check parameters */
-    if (conn==NULL || ((clientin==NULL) && (clientinlen>0)))
-	return SASL_BADPARAM;
+    if (!conn) return SASL_BADPARAM;
+    if ((clientin==NULL) && (clientinlen>0))
+	PARAMERROR(conn);
 
     ret = s_conn->mech->plug->mech_step(conn->context,
 					s_conn->sparams,
@@ -1098,7 +1160,7 @@ int sasl_server_step(sasl_conn_t *conn,
 	conn->oparams.maxoutbuf = DEFAULT_MAXOUTBUF;
     }
 
-    return ret;
+    RETURN(conn, ret);
 }
 
 /* returns the length of all the mechanisms
@@ -1158,9 +1220,10 @@ int sasl_listmech(sasl_conn_t *conn,
 
   /* if there hasn't been a sasl_sever_init() fail */
   if (_sasl_server_active==0) return SASL_NOTINIT;
-
-  if (! conn || ! result)
-    return SASL_BADPARAM;
+  if (!conn) return SASL_BADPARAM;
+  
+  if (! result)
+      PARAMERROR(conn);
 
   if (plen != NULL)
       *plen = 0;
@@ -1173,19 +1236,17 @@ int sasl_listmech(sasl_conn_t *conn,
       mysep = " ";
   }
 
-  if (! mechlist)
-    return SASL_FAIL;
-
-  if (mechlist->mech_length <= 0)
-    return SASL_NOMECH;
+  if (! mechlist || mechlist->mech_length <= 0)
+      INTERROR(conn, SASL_NOMECH);
 
   resultlen = (prefix ? strlen(prefix) : 0)
             + (strlen(mysep) * (mechlist->mech_length - 1))
 	    + mech_names_len()
             + (suffix ? strlen(suffix) : 0)
 	    + 1;
-  ret = _buf_alloc(&s_conn->mechlist_buf, &s_conn->mechlist_buf_len, resultlen);
-  if(ret != SASL_OK) return ret;
+  ret = _buf_alloc(&s_conn->mechlist_buf,
+		   &s_conn->mechlist_buf_len, resultlen);
+  if(ret != SASL_OK) MEMERROR(conn);
 
   if (prefix)
     strcpy (s_conn->mechlist_buf,prefix);
@@ -1240,12 +1301,23 @@ static int _sasl_checkpass(sasl_conn_t *conn, const char *service,
 			   const char *user, const char *pass)
 {
     sasl_server_conn_t *s_conn = (sasl_server_conn_t *) conn;
-    int result = SASL_NOMECH;
+    int result;
     sasl_getopt_t *getopt;
+    sasl_server_userdb_checkpass_t *checkpass_cb;
     void *context;
     const char *mech;
     struct sasl_verify_password_s *v;
 
+    /* call userdb callback function, if available */
+    result = _sasl_getcallback(conn, SASL_CB_SERVER_USERDB_CHECKPASS,
+			       &checkpass_cb, &context);
+    if(result == SASL_OK && checkpass_cb) {
+	result = checkpass_cb(conn, context, user, pass, strlen(pass),
+			      s_conn->sparams->propctx);
+	if(result == SASL_OK)
+	    return SASL_OK;
+    }
+
     /* figure out how to check (i.e. sasldb or saslauthd or pwcheck) */
     if (_sasl_getcallback(conn, SASL_CB_GETOPT, &getopt, &context)
             == SASL_OK) {
@@ -1254,6 +1326,8 @@ static int _sasl_checkpass(sasl_conn_t *conn, const char *service,
 
     if(!mech) mech = DEFAULT_CHECKPASS_MECH;
 
+    result = SASL_NOMECH;
+
     for (v = _sasl_verify_password; v->name; v++) {
 	if(is_mech(mech, v->name)) {
 	    result = v->verify(conn, user, pass, service, s_conn->user_realm);
@@ -1266,23 +1340,24 @@ static int _sasl_checkpass(sasl_conn_t *conn, const char *service,
 	_sasl_log(conn, SASL_LOG_ERR, "unkwnown password verifier %s", mech);
     }
 
-    return result;
+    if (result != SASL_OK)
+	sasl_seterror(conn, SASL_NOLOG, "checkpass failed");
+
+    RETURN(conn, result);
 }
 
-
 /* check if a plaintext password is valid
- * if user is NULL, check if plaintext is enabled
+ *   if user is NULL, check if plaintext passwords are enabled
  * inputs:
- *  user         -- user to query in current user_realm
- *  userlen      -- length of username, 0 = strlen(user)
- *  pass         -- plaintext password to check
- *  passlen      -- length of password, 0 = strlen(pass)
- * outputs:
- *  errstr       -- set to error message for use in protocols
+ *  user          -- user to query in current user_domain
+ *  userlen       -- length of username, 0 = strlen(user)
+ *  pass          -- plaintext password to check
+ *  passlen       -- length of password, 0 = strlen(pass)
  * returns 
- *  SASL_OK      -- success
- *  SASL_NOMECH  -- user found, but no verifier
- *  SASL_NOUSER  -- user not found
+ *  SASL_OK       -- success
+ *  SASL_NOMECH   -- mechanism not supported
+ *  SASL_NOVERIFY -- user found, but no verifier
+ *  SASL_NOUSER   -- user not found
  */
 int sasl_checkpass(sasl_conn_t *conn,
 		   const char *user,
@@ -1292,11 +1367,17 @@ int sasl_checkpass(sasl_conn_t *conn,
 {
     int result;
 
-    /* check params */
     if (_sasl_server_active==0) return SASL_NOTINIT;
     
-    if ((conn == NULL) || (user == NULL) || (pass == NULL))
-	return SASL_BADPARAM;
+    /* check if it's just a query if we are enabled */
+    if (!user)
+	return SASL_OK;
+
+    if (!conn) return SASL_BADPARAM;
+    
+    /* check params */
+    if (pass == NULL)
+	PARAMERROR(conn);
 
     result = _sasl_checkpass(conn, conn->service, user, pass);
 
@@ -1307,7 +1388,7 @@ int sasl_checkpass(sasl_conn_t *conn,
 	result = _sasl_transition(conn, pass, passlen);
     }
 
-    return result;
+    RETURN(conn,result);
 }
 
 /* check if a user exists on server
@@ -1333,8 +1414,12 @@ int sasl_user_exists(sasl_conn_t *conn,
 
     /* check params */
     if (_sasl_server_active==0) return SASL_NOTINIT;
+    if (!conn) return SASL_BADPARAM;
+    
+    if(!user) 
+	PARAMERROR(conn);
 
-    if(!conn || !service || !user) return SASL_BADPARAM;
+    if(!service) service = conn->service;
 
     for (v = _sasl_verify_password; v->name; v++) {
 	result = v->verify(conn, user, NULL, service, user_realm);
@@ -1350,65 +1435,87 @@ int sasl_user_exists(sasl_conn_t *conn,
     if (result == SASL_NOMECH) {
 	/* no mechanism available ?!? */
 	_sasl_log(conn, SASL_LOG_ERR, "no plaintext password verifier?");
+	sasl_seterror(conn, SASL_NOLOG, "no plaintext password verifier?");
     }
-    
-    return result;
+
+    RETURN(conn, result);
 }
 
-
-/* check if an APOP exchange is valid
+/* check if an apop exchange is valid
  *  (note this is an optional part of the SASL API)
- *  if challenge is NULL, just check if APOP is installed
+ *  if challenge is NULL, just check if APOP is enabled
  * inputs:
  *  challenge     -- challenge which was sent to client
  *  challen       -- length of challenge, 0 = strlen(challenge)
- *  response      -- client response string: "32HEXDIGIT"
+ *  response      -- client response, "<user> <digest>" (RFC 1939)
  *  resplen       -- length of response, 0 = strlen(response)
  * returns 
  *  SASL_OK       -- success
- *  SASL_FAIL     -- failure
- *  SASL_BADPARAM -- invalid parameter
  *  SASL_BADAUTH  -- authentication failed
+ *  SASL_BADPARAM -- missing challenge
+ *  SASL_BADPROT  -- protocol error (e.g., response in wrong format)
+ *  SASL_NOVERIFY -- user found, but no verifier
+ *  SASL_NOMECH   -- mechanism not supported
  *  SASL_NOUSER   -- user not found
  */
 int sasl_checkapop(sasl_conn_t *conn,
+#ifdef DO_SASL_CHECKAPOP
  		   const char *challenge,
  		   unsigned challen __attribute__((unused)),
  		   const char *response,
  		   unsigned resplen __attribute__((unused)))
+#else
+ 		   const char *challenge __attribute__((unused)),
+ 		   unsigned challen __attribute__((unused)),
+ 		   const char *response __attribute__((unused)),
+ 		   unsigned resplen __attribute__((unused)))
+#endif
 {
+#ifdef DO_SASL_CHECKAPOP
     sasl_server_conn_t *s_conn = (sasl_server_conn_t *) conn;
     char *user, *user_end;
     size_t user_len;
-    int ret;
-    int result = SASL_FAIL;
+    int result;
 
-    if (_sasl_server_active==0) return SASL_NOTINIT;
+    if (_sasl_server_active==0)
+	return SASL_NOTINIT;
 
-    /* check if it's just a query if we are installed */
+    /* check if it's just a query if we are enabled */
     if(!challenge)
 	return SASL_OK;
 
-    /* check other params */
-    if (!conn || !response)
-	return SASL_BADPARAM;
+    /* check params */
+    if (!conn) return SASL_BADPARAM;
+    if (!response)
+	PARAMERROR(conn);
 
-    /* Parse out username */
+    /* Parse out username and digest.
+     *
+     * Per RFC 1939, response must be "<user> <digest>", where
+     * <digest> is a 16-octet value which is sent in hexadecimal
+     * format, using lower-case ASCII characters.
+     */
     user_end = strrchr(response, ' ');
+    if (!user_end || strspn(user_end + 1, "0123456789abcdef") != 32) 
+    {
+        sasl_seterror(conn, 0, "Bad Digest");
+        RETURN(conn,SASL_BADPROT);
+    }
+ 
     user_len = (size_t)(user_end - response);
     user = sasl_ALLOC(user_len + 1);
     memcpy(user, response, user_len);
     user[user_len] = '\0';
 
     /* Cannonify it */
-    ret = _sasl_canon_user(conn,
+    result = _sasl_canon_user(conn,
 			   user, user_len,
 			   user, user_len,
 			   0, &(conn->oparams));
 
     sasl_FREE(user);
 
-    if(ret != SASL_OK) return ret;
+    if(result != SASL_OK) RETURN(conn, result);
 
     /* Do APOP verification */
     result = _sasl_sasldb_verify_apop(conn, conn->oparams.user,
@@ -1421,6 +1528,11 @@ int sasl_checkapop(sasl_conn_t *conn,
 	conn->oparams.authid = NULL;
     }
 
-    return result;
+    RETURN(conn, result);
+#else /* sasl_checkapop was disabled at compile time */
+    sasl_seterror(conn, SASL_NOLOG,
+	"sasl_checkapop called, but was disabled at compile time");
+    RETURN(conn, SASL_NOMECH);
+#endif /* DO_SASL_CHECKAPOP */
 }
  
