@@ -1,7 +1,7 @@
 /* GSSAPI SASL plugin
  * Leif Johansson
  * Rob Siemborski (SASL v2 Conversion)
- * $Id: gssapi.c,v 1.41.2.13 2001/06/27 14:56:30 rjs3 Exp $
+ * $Id: gssapi.c,v 1.41.2.21 2001/07/12 22:05:32 rjs3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -121,8 +121,10 @@ typedef struct context {
 
     char *encode_buf;                /* For encoding/decoding mem management */
     char *decode_buf;
+    char *decode_once_buf;
     unsigned encode_buf_len;
     unsigned decode_buf_len;
+    unsigned decode_once_buf_len;
     buffer_info_t *enc_in_buf;
 
     char *out_buf;                   /* per-step mem management */
@@ -180,6 +182,7 @@ sasl_gss_encode(void *context, const struct iovec *invec, unsigned numiov,
     {
       if (output_token->value)
 	  gss_release_buffer(&min_stat, output_token);
+      SETERROR(text->utils, "GSSAPI Failure");
       return SASL_FAIL;
     }
 
@@ -242,8 +245,10 @@ sasl_gss_decode_once(void *context,
     int result;
     unsigned diff;
 
-    if (text->state != SASL_GSSAPI_STATE_AUTHENTICATED)
+    if (text->state != SASL_GSSAPI_STATE_AUTHENTICATED) {
+	SETERROR(text->utils, "GSSAPI Failure");
 	return SASL_NOTDONE;
+    }
 
     /* first we need to extract a packet */
     if (text->needsize > 0) {
@@ -261,14 +266,16 @@ sasl_gss_decode_once(void *context,
 	    text->size = ntohl(text->size);
 	    text->cursize = 0;
 
-	    if (text->size > 0xFFFF || text->size <= 0) return SASL_FAIL;
-
+	    if (text->size > 0xFFFF || text->size <= 0) {
+		SETERROR(text->utils, "GSSAPI Failure");
+		return SASL_FAIL;
+	    }
+	    
 	    if (text->bufsize < text->size + 5) {
 		result = _plug_buf_alloc(text->utils, &text->buffer,
 					       &(text->bufsize), text->size+5);
 		if(result != SASL_OK) return result;
 	    }
-	    if (text->buffer == NULL) return SASL_NOMEM;
 	}
 	if (*inputlen == 0) {
 	    /* need more data ! */
@@ -314,18 +321,26 @@ sasl_gss_decode_once(void *context,
     {
 	if (output_token->value)
 	    gss_release_buffer(&min_stat, output_token);
+	SETERROR(text->utils, "GSSAPI Failure");
 	return SASL_FAIL;
     }
 
     if (outputlen)
 	*outputlen = output_token->length;
 
-    /* FIXME: This really looks like a potential memory leak to me! */
     if (output_token->value) {
-	if (output)
-	    *output = output_token->value;
-	else
-	    gss_release_buffer(&min_stat, output_token);
+	if (output) {
+	    result = _plug_buf_alloc(text->utils, &text->decode_once_buf,
+				     &text->decode_once_buf_len,
+				     *outputlen);
+	    if(result != SASL_OK) {
+		gss_release_buffer(&min_stat, output_token);
+		return result;
+	    }
+	    *output = text->decode_once_buf;
+	    memcpy(*output, output_token->value, *outputlen);
+	}
+	gss_release_buffer(&min_stat, output_token);
     }
 
     /* reset for the next packet */
@@ -348,6 +363,7 @@ static int sasl_gss_decode(void *context,
 
     while (inputlen!=0)
     {
+	/* no need to free tmp */
       ret = sasl_gss_decode_once(text, &input, &inputlen,
 				 &tmp, &tmplen);
 
@@ -367,7 +383,6 @@ static int sasl_gss_decode(void *context,
 	  *(text->decode_buf + *outputlen + tmplen) = '\0';
 
 	  *outputlen+=tmplen;
-	  text->utils->free(tmp);
       }
     }
 
@@ -399,8 +414,10 @@ sasl_gss_server_start(void *glob_context __attribute__((unused)),
   context_t *text;
   
   text = gss_new_context(params->utils);
-  if (text == NULL) 
-    return SASL_NOMEM;
+  if (text == NULL) {
+      MEMERROR(params->utils);
+      return SASL_NOMEM;
+  }
 
   text->gss_ctx = GSS_C_NO_CONTEXT;
   text->client_name = GSS_C_NO_NAME;
@@ -453,6 +470,11 @@ sasl_gss_free_context_contents(context_t *text)
       text->decode_buf = NULL;
   }
 
+  if (text->decode_once_buf) {
+      text->utils->free(text->decode_once_buf);
+      text->decode_once_buf = NULL;
+  }
+  
   if (text->enc_in_buf) {
       if(text->enc_in_buf->data) text->utils->free(text->enc_in_buf->data);
       text->utils->free(text->enc_in_buf);
@@ -504,7 +526,10 @@ sasl_gss_server_step (void *conn_context,
   output_token->value = NULL; output_token->length = 0;
   input_token->value = NULL; input_token->length = 0;
 
-  if(!serverout) return SASL_BADPARAM;
+  if(!serverout) {
+      PARAMERROR(text->utils);
+      return SASL_BADPARAM;
+  }
 
   switch (text->state)
     {
@@ -525,8 +550,9 @@ sasl_gss_server_step (void *conn_context,
 	  name_token.value = (char *)params->utils->malloc((name_token.length + 1) * sizeof(char));
 	  if (name_token.value == NULL)
 	    {
-	      sasl_gss_free_context_contents(text);
-	      return SASL_NOMEM;
+		MEMERROR(text->utils);
+		sasl_gss_free_context_contents(text);
+		return SASL_NOMEM;
 	    }
 	  sprintf(name_token.value,"%s@%s", params->service, params->serverFQDN);
 
@@ -539,6 +565,7 @@ sasl_gss_server_step (void *conn_context,
 	  name_token.value = NULL;
 	  
 	  if (GSS_ERROR(maj_stat)) {
+	      SETERROR(text->utils, "GSSAPI Failure");
 	      sasl_gss_free_context_contents(text);
 	      return SASL_FAIL;
 	  }
@@ -558,6 +585,7 @@ sasl_gss_server_step (void *conn_context,
 				      NULL);
 	  
 	  if (GSS_ERROR(maj_stat)) {
+	      SETERROR(text->utils, "GSSAPI Failure");
 	      sasl_gss_free_context_contents(text);
 	      return SASL_FAIL;
 	  }
@@ -569,8 +597,6 @@ sasl_gss_server_step (void *conn_context,
 	  real_input_token.length = clientinlen;
 	}
       
-      VL(("sasl_gss_server_step: AUTHNEG\n"));
-
       maj_stat =
 	gss_accept_sec_context (&min_stat,
 				&(text->gss_ctx),
@@ -588,7 +614,7 @@ sasl_gss_server_step (void *conn_context,
 	  if (output_token->value) {
 	      gss_release_buffer(&min_stat, output_token);
 	  }
-
+	  SETERROR(text->utils, "GSSAPI Failure");
 	  sasl_gss_free_context_contents(text);
 	  return SASL_BADAUTH;
       }
@@ -613,7 +639,6 @@ sasl_gss_server_step (void *conn_context,
       
       if (maj_stat == GSS_S_COMPLETE)
 	{
-	  VL (("GSS_S_COMPLETE\n"));
           /* Switch to ssf negotiation */
 	  text->state = SASL_GSSAPI_STATE_SSFCAP;
 	}
@@ -630,7 +655,6 @@ sasl_gss_server_step (void *conn_context,
 
 	name_token.value = NULL;
 	name_without_realm.value = NULL;
-	VL(("sasl_gss_server_step: SSFCAP\n"));
 
 	name_token.value = NULL;
 	
@@ -649,6 +673,7 @@ sasl_gss_server_step (void *conn_context,
 		gss_release_buffer(&min_stat, &name_token);
 	    if (without)
 		gss_release_name(&min_stat, &without);
+	    SETERROR(text->utils, "GSSAPI Failure");
 	    sasl_gss_free_context_contents(text);
 	    return SASL_BADAUTH;
 	}
@@ -662,7 +687,10 @@ sasl_gss_server_step (void *conn_context,
 	    /* NOTE: libc malloc, as it is freed below by a gssapi internal
 	     *       function! */
 	    name_without_realm.value = malloc(strlen(name_token.value)+1);
-	    if (name_without_realm.value == NULL) return SASL_NOMEM;
+	    if (name_without_realm.value == NULL) {
+		MEMERROR(text->utils);
+		return SASL_NOMEM;
+	    }
 
 	    strcpy(name_without_realm.value, name_token.value);
 
@@ -682,6 +710,7 @@ sasl_gss_server_step (void *conn_context,
 		    gss_release_buffer(&min_stat, &name_token);
 		if (without)
 		    gss_release_name(&min_stat, &without);
+		SETERROR(text->utils, "GSSAPI Failure");
 		sasl_gss_free_context_contents(text);
 		return SASL_BADAUTH;
 	    }
@@ -697,6 +726,7 @@ sasl_gss_server_step (void *conn_context,
 		    gss_release_buffer(&min_stat, &name_token);
 		if (without)
 		    gss_release_name(&min_stat, &without);
+		SETERROR(text->utils, "GSSAPI Failure");
 		sasl_gss_free_context_contents(text);
 		return SASL_BADAUTH;
 	    }
@@ -710,12 +740,20 @@ sasl_gss_server_step (void *conn_context,
 	{
 	    text->u.authid =
 		(char *)params->utils->malloc(strlen(name_without_realm.value)+1);
-	    if (text->u.authid == NULL) return SASL_NOMEM;
+	    if (text->u.authid == NULL) {
+		MEMERROR(params->utils);
+		return SASL_NOMEM;
+	    }
+	    
 	    strcpy(text->u.authid, name_without_realm.value);
 	} else {
 	    text->u.authid =
 		(char *)params->utils->malloc(strlen(name_token.value)+1);
-	    if (text->u.authid == NULL) return SASL_NOMEM;
+	    if (text->u.authid == NULL) {
+		MEMERROR(params->utils);
+		return SASL_NOMEM;
+	    }
+	    
 	    strcpy(text->u.authid, name_token.value);
 	}
 
@@ -767,6 +805,7 @@ sasl_gss_server_step (void *conn_context,
 	if (GSS_ERROR(maj_stat)) {
 	    if (output_token->value)
 		gss_release_buffer(&min_stat, output_token);
+	    SETERROR(text->utils, "GSSAPI Failure");
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	}
@@ -789,7 +828,6 @@ sasl_gss_server_step (void *conn_context,
 	  gss_release_buffer(&min_stat, output_token);
         }
 
-	VL(("Sending %d bytes (ssfcap) to client\n",*serveroutlen));
         /* Wait for ssf request and authid */
 	text->state = SASL_GSSAPI_STATE_SSFREQ; 
 	
@@ -812,6 +850,7 @@ sasl_gss_server_step (void *conn_context,
 			       NULL);
 	
 	if (GSS_ERROR(maj_stat)) {
+	    SETERROR(text->utils, "GSSAPI Failure");
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	}
@@ -837,12 +876,11 @@ sasl_gss_server_step (void *conn_context,
 		    "protocol violation: client requested invalid layer");
 	    if (output_token->value)
 		gss_release_buffer(&min_stat, output_token);
+	    SETERROR(text->utils, "GSSAPI Failure");
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	}
 
-	VL (("Got %d bytes from client\n",output_token->length));
-
 	if (output_token->length > 4) {
 	    int ret;
 
@@ -855,11 +893,9 @@ sasl_gss_server_step (void *conn_context,
 
 	    if (ret != SASL_OK) {
 		sasl_gss_free_context_contents(text);
-		return SASL_NOMEM;
+		return ret;
 	    }
 
-	    VL(("Got user %s\n",oparams->user));
-
 	    memcpy(&oparams->maxoutbuf,((char *) real_output_token.value) + 1,
 		   sizeof(unsigned));
 	    oparams->maxoutbuf = ntohl(oparams->maxoutbuf);
@@ -882,11 +918,13 @@ sasl_gss_server_step (void *conn_context,
       break;
 
     default:
-      return SASL_FAIL;
-      break;
+	SETERROR(text->utils, "GSSAPI Failure");
+	return SASL_FAIL;
+	break;
     }
 
   /* we should never get here ! */
+  SETERROR(text->utils, "GSSAPI Failure");
   return SASL_FAIL;
 }
 
@@ -910,8 +948,7 @@ static sasl_server_plug_t plugins[] =
   }
 };
 
-int 
-sasl_server_plug_init(
+int gssapiv2_server_plug_init(
 #ifndef HAVE_GSSKRB5_REGISTER_ACCEPTOR_IDENTITY
 		      const sasl_utils_t *utils __attribute__((unused)),
 #else
@@ -920,11 +957,11 @@ sasl_server_plug_init(
 		      int maxversion,
 		      int *out_version,
 		      sasl_server_plug_t **pluglist,
-		      int *plugcount,
-		      const char *plugname __attribute__((unused)))
+		      int *plugcount)
 {
 #ifdef HAVE_GSSKRB5_REGISTER_ACCEPTOR_IDENTITY
     const char *keytab;
+    char keytab_path[1024];
     unsigned int rl;
 #endif
 
@@ -943,10 +980,19 @@ sasl_server_plug_init(
 	if (access(keytab, R_OK) != 0) {
 	    utils->log(NULL, SASL_LOG_ERR,
 		       "can't access keytab file %s: %m", keytab, errno);
+	    SETERROR(utils, "GSSAPI Failure");
 	    return SASL_FAIL;
 	}
+
+	if(strlen(keytab) > 1024) {
+	    utils->log(NULL, SASL_LOG_ERR,
+		       "path to keytab is > 1024 characters");
+	    return SASL_BUFOVER;
+	}
 	
-	gsskrb5_register_acceptor_identity(keytab);
+	strncpy(keytab_path, keytab, 1024);
+	
+	gsskrb5_register_acceptor_identity(keytab_path);
     }
 #endif
 
@@ -993,8 +1039,11 @@ make_prompts(sasl_client_params_t *params,
    if (auth_res==SASL_INTERACT) num++;
    if (pass_res==SASL_INTERACT) num++;
  
-   if (num==1) return SASL_FAIL;
- 
+   if (num==1) {
+       SETERROR(params->utils, "GSSAPI Failure");
+       return SASL_FAIL;
+   }
+   
    prompts=params->utils->malloc(sizeof(sasl_interact_t)*(num+1));
    if ((prompts) ==NULL) return SASL_NOMEM;
    *prompts_res=prompts;
@@ -1131,11 +1180,6 @@ sasl_gss_client_step (void *conn_context,
   input_token->value = NULL; 
   input_token->length = 0;
 
-  if (!clientout && text->state == SASL_GSSAPI_STATE_AUTHNEG) {
-      /* initial client send not allowed */
-      return SASL_CONTINUE;
-  }
-
   *clientout = NULL;
   *clientoutlen = 0;
 
@@ -1143,22 +1187,15 @@ sasl_gss_client_step (void *conn_context,
     {
     case SASL_GSSAPI_STATE_AUTHNEG:
       {
-	  VL(("sasl_gss_client_step: AUTHNEG\n"));
-
 	  /* try to get the userid */
 	  if (text->u.user ==NULL)
 	  {
 	    int auth_result = SASL_OK;
-	    VL (("Trying to get userid\n"));
 
 	    auth_result=get_userid(params,
 				   &text->u.user,
 				   prompt_need);
 
-	    if (text->u.user) {
-		VL (("Userid: %s\n",text->u.user));
-	    }
-
 	    if ((auth_result!=SASL_OK) && (auth_result!=SASL_INTERACT))
 	      {
 		sasl_gss_free_context_contents(text);
@@ -1166,7 +1203,7 @@ sasl_gss_client_step (void *conn_context,
 	      }
 
 	    /* free prompts we got */
-	    if (prompt_need) {
+	    if (prompt_need && *prompt_need) {
 		params->utils->free(*prompt_need);
 		*prompt_need = NULL;
 	    }
@@ -1201,10 +1238,13 @@ sasl_gss_client_step (void *conn_context,
 		sasl_gss_free_context_contents(text);
 		return SASL_NOMEM;
 	      }
-	    if (params->serverFQDN == NULL || strlen(params->serverFQDN) == 0)
-	      return SASL_FAIL;
+	    if (params->serverFQDN == NULL
+		|| strlen(params->serverFQDN) == 0) {
+		SETERROR(text->utils, "GSSAPI Failure");
+		return SASL_FAIL;
+	    }
+	    
 	    sprintf(name_token.value,"%s@%s", params->service, params->serverFQDN);
-	    VL(("name: %s\n",(char *)name_token.value)); /* */
 
 	    maj_stat = gss_import_name (&min_stat,
 					&name_token,
@@ -1215,10 +1255,11 @@ sasl_gss_client_step (void *conn_context,
 	    name_token.value = NULL;
 	    
 	    if (GSS_ERROR(maj_stat))
-	      {
+	    {
+		SETERROR(text->utils, "GSSAPI Failure");
 		sasl_gss_free_context_contents(text);
 		return SASL_FAIL;
-	      }
+	    }
 	  }
 
 	if (serverinlen)
@@ -1255,6 +1296,7 @@ sasl_gss_client_step (void *conn_context,
 	{
 	    if (output_token->value)
 		gss_release_buffer(&min_stat, output_token);
+	    SETERROR(text->utils, "GSSAPI Failure");
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	}
@@ -1277,10 +1319,10 @@ sasl_gss_client_step (void *conn_context,
         }
 
 	if (maj_stat == GSS_S_COMPLETE)
-	  {
-	    VL(("GSS_S_COMPLETE\n"));
-	    text->state = SASL_GSSAPI_STATE_SSFCAP; /* Switch to ssf negotiation */
-	  }
+	{
+            /* Switch to ssf negotiation */
+	    text->state = SASL_GSSAPI_STATE_SSFCAP;
+	}
 	
 	return SASL_CONTINUE;
 	break;
@@ -1292,8 +1334,6 @@ sasl_gss_client_step (void *conn_context,
 	sasl_ssf_t need, allowed;
 	char serverhas, mychoice;
 
-	VL(("sasl_gss_client_step: SSFCAP\n"));
-
 	real_input_token.value = (void *) serverin;
 	real_input_token.length = serverinlen;
 
@@ -1305,6 +1345,7 @@ sasl_gss_client_step (void *conn_context,
 			       NULL);
 	
 	if (GSS_ERROR(maj_stat)) {
+	    SETERROR(text->utils, "GSSAPI Failure");
 	    sasl_gss_free_context_contents(text);
 	    if (output_token->value)
 		gss_release_buffer(&min_stat, output_token);
@@ -1341,21 +1382,18 @@ sasl_gss_client_step (void *conn_context,
 	    oparams->decode = &sasl_gss_decode;
 	    oparams->mech_ssf = 56;
 	    mychoice = 4;
-	    VL (("Using encryption layer\n"));
 	} else if (allowed >= 1 && need <= 1 && (serverhas & 2)) {
 	    /* integrity */
 	    oparams->encode = &sasl_gss_integrity_encode;
 	    oparams->decode = &sasl_gss_decode;
 	    oparams->mech_ssf = 1;
 	    mychoice = 2;
-	    VL (("Using integrity layer\n"));
 	} else if (need <= 0 && (serverhas & 1)) {
 	    /* no layer */
 	    oparams->encode = NULL;
 	    oparams->decode = NULL;
 	    oparams->mech_ssf = 0;
 	    mychoice = 1;
-	    VL (("Using no layer\n"));
 	} else {
 	    /* there's no appropriate layering for us! */
 	    sasl_gss_free_context_contents(text);
@@ -1376,7 +1414,7 @@ sasl_gss_client_step (void *conn_context,
 	    sasl_gss_free_context_contents(text);
 	    return SASL_NOMEM;
 	  }
-	VL(("user: %s,buflen=%d\n",oparams->user,input_token->length));
+
 	if (oparams->user)
 	    memcpy((char *)input_token->value+4,oparams->user,alen);
 	
@@ -1401,6 +1439,7 @@ sasl_gss_client_step (void *conn_context,
 	  {
 	    if (output_token->value)
 		gss_release_buffer(&min_stat, output_token);
+	    SETERROR(text->utils, "GSSAPI Failure");
 	    sasl_gss_free_context_contents(text);
 	    return SASL_FAIL;
 	  }
@@ -1427,16 +1466,17 @@ sasl_gss_client_step (void *conn_context,
 	return SASL_OK;
       }
     case SASL_GSSAPI_STATE_AUTHENTICATED:
-      VL(("sasl_gss_client_step: AUTHENTICATED\n"));
       return SASL_OK;
       break;
 
     default:
-      return SASL_FAIL;
-      break;
+	SETERROR(text->utils, "GSSAPI Failure");
+	return SASL_FAIL;
+	break;
     }
 
   /* we should never get here!!! */
+  SETERROR(text->utils, "GSSAPI Failure");
   return SASL_FAIL;
 }
 
@@ -1459,13 +1499,12 @@ static sasl_client_plug_t client_plugins[] =
   }
 };
 
-int 
-sasl_client_plug_init(const sasl_utils_t *utils __attribute__((unused)), 
-		      int maxversion,
-		      int *out_version, 
-		      sasl_client_plug_t **pluglist,
-		      int *plugcount,
-		      const char *plugname __attribute__((unused)))
+int gssapiv2_client_plug_init(
+    const sasl_utils_t *utils __attribute__((unused)), 
+    int maxversion,
+    int *out_version, 
+    sasl_client_plug_t **pluglist,
+    int *plugcount)
 {
   if (maxversion<SASL_CLIENT_PLUG_VERSION)
     return SASL_BADVERS;
