@@ -1,7 +1,7 @@
 /* Kerberos4 SASL plugin
  * Rob Siemborski
  * Tim Martin 
- * $Id: kerberos4.c,v 1.65.2.20 2001/06/28 19:28:53 rjs3 Exp $
+ * $Id: kerberos4.c,v 1.65.2.28 2001/07/12 17:42:55 rjs3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -122,6 +122,12 @@ extern int gethostname(char *, int);
 #define KRB_DES_SECURITY_BITS (56)
 #define KRB_INTEGRITY_BITS (1)
 
+typedef enum Krb_sec {
+    KRB_SEC_NONE = 0,
+    KRB_SEC_INTEGRITY = 1,
+    KRB_SEC_ENCRYPTION = 2
+} Krb_sec_t;
+
 typedef struct context {
   int state;
 
@@ -154,10 +160,13 @@ typedef struct context {
 
   const sasl_utils_t *utils;       /* this is useful to have around */
 
+  Krb_sec_t sec_type;
   char *encode_buf;                /* For encoding/decoding mem management */
   char *decode_buf;
+  char *decode_once_buf;
   unsigned encode_buf_len;
   unsigned decode_buf_len;
+  unsigned decode_once_buf_len;
   buffer_info_t *enc_in_buf;
 
   char *out_buf;                   /* per-step mem management */
@@ -196,11 +205,11 @@ static void *krb_mutex = NULL;
 static char *srvtab = NULL;
 static unsigned refcount = 0;
 
-static int privacy_encode(void *context,
-			  const struct iovec *invec,
-			  unsigned numiov,
-			  const char **output,
-			  unsigned *outputlen)
+static int encode(void *context,
+		  const struct iovec *invec,
+		  unsigned numiov,
+		  const char **output,
+		  unsigned *outputlen)
 {
   int len, ret;
   context_t *text = (context_t *)context;
@@ -215,11 +224,19 @@ static int privacy_encode(void *context,
 
   KRB_LOCK_MUTEX(text->utils);
   
-  len=krb_mk_priv((char *) text->enc_in_buf->data, text->encode_buf+4,
-		  text->enc_in_buf->curlen,  text->init_keysched, 
-		  &text->session, &text->ip_local,
-		  &text->ip_remote);
-
+  if(text->sec_type == KRB_SEC_ENCRYPTION) {
+      len=krb_mk_priv(text->enc_in_buf->data, (text->encode_buf+4),
+		      text->enc_in_buf->curlen,  text->init_keysched, 
+		      &text->session, &text->ip_local,
+		      &text->ip_remote);
+  } else if (text->sec_type == KRB_SEC_INTEGRITY) {
+      len=krb_mk_safe(text->enc_in_buf->data, (text->encode_buf+4),
+		      text->enc_in_buf->curlen,
+		      &text->session, &text->ip_local, &text->ip_remote);
+  } else {
+      len = -1;
+  }
+  
   KRB_UNLOCK_MUTEX(text->utils);
   
   /* returns -1 on error */
@@ -237,9 +254,9 @@ static int privacy_encode(void *context,
 }
 
 
-static int privacy_decode_once(void *context,
-			       const char **input, unsigned *inputlen,
-			       char **output, unsigned *outputlen)
+static int decode_once(void *context,
+		       const char **input, unsigned *inputlen,
+		       char **output, unsigned *outputlen)
 {
     int tocopy, result;
     unsigned diff;
@@ -306,9 +323,20 @@ static int privacy_decode_once(void *context,
 
     KRB_LOCK_MUTEX(text->utils);
     
-    result=krb_rd_priv((char *) text->buffer,text->size,  text->init_keysched, 
-		       &text->session, &text->ip_remote,
-		       &text->ip_local, &data);
+    if(text->sec_type == KRB_SEC_ENCRYPTION) {
+	result=krb_rd_priv(text->buffer,text->size, text->init_keysched, 
+			   &text->session, &text->ip_remote,
+			   &text->ip_local, &data);
+    } else if (text->sec_type == KRB_SEC_INTEGRITY) {
+        result = krb_rd_safe(text->buffer, text->size,
+			     &text->session, &text->ip_remote,
+			     &text->ip_local, &data);
+    } else {
+        KRB_UNLOCK_MUTEX(text->utils);
+	text->utils->seterror(text->utils->conn, 0,
+			      "KERBEROS_4 decode called with KRB_SEC_NONE");
+	return SASL_FAIL;
+    }
 
     KRB_UNLOCK_MUTEX(text->utils);
 
@@ -330,12 +358,13 @@ static int privacy_decode_once(void *context,
     text->time_sec = data.time_sec;
     text->time_5ms = data.time_5ms;
 
-    *output = text->utils->malloc(data.app_length + 1);
-
-    if ((*output) == NULL) {
-	return SASL_NOMEM;
-    }
+    result = _plug_buf_alloc(text->utils, &text->decode_once_buf,
+			      &text->decode_once_buf_len,
+			      data.app_length + 1);
+    if(result != SASL_OK)
+	return result;
     
+    *output = text->decode_once_buf;
     *outputlen = data.app_length;
     memcpy(*output, data.app_data, data.app_length);
     (*output)[*outputlen] = '\0';
@@ -346,9 +375,9 @@ static int privacy_decode_once(void *context,
     return SASL_OK;
 }
 
-static int privacy_decode(void *context,
-			  const char *input, unsigned inputlen,
-			  const char **output, unsigned *outputlen)
+static int decode(void *context,
+		  const char *input, unsigned inputlen,
+		  const char **output, unsigned *outputlen)
 {
     char *tmp = NULL;
     unsigned tmplen = 0;
@@ -359,15 +388,15 @@ static int privacy_decode(void *context,
 
     while (inputlen!=0)
     {
-      ret = privacy_decode_once(text, &input, &inputlen,
-				&tmp, &tmplen);
-
+      /* No need to free tmp, it will be reused */
+      ret = decode_once(text, &input, &inputlen, &tmp, &tmplen);
       if(ret != SASL_OK) return ret;
 
       if (tmp!=NULL) /* if received 2 packets merge them together */
       {
 	  ret = _plug_buf_alloc(text->utils, &text->decode_buf,
-				&text->decode_buf_len, *outputlen + tmplen + 1);
+				&text->decode_buf_len,
+				*outputlen + tmplen + 1);
 	  if(ret != SASL_OK) return ret;
 
 	  *output = text->decode_buf;
@@ -377,199 +406,20 @@ static int privacy_decode(void *context,
 	  *(text->decode_buf + *outputlen + tmplen) = '\0';	  
 
 	  *outputlen+=tmplen;
-	  text->utils->free(tmp);
       }
     }
 
     return SASL_OK;
 }
 
-static int
-integrity_encode(void *context,
-		 const struct iovec *invec,
-		 unsigned numiov,
-		 const char **output,
-		 unsigned *outputlen)
-{
-  int len, ret;
-  context_t *text;
-  text=context;
-
-  ret = _plug_iovec_to_buf(text->utils, invec, numiov, &text->enc_in_buf);
-  if(ret != SASL_OK) return ret;
-
-  ret = _plug_buf_alloc(text->utils, &text->encode_buf, &text->encode_buf_len,
-			text->enc_in_buf->curlen+40);
-
-  if(ret != SASL_OK) return ret;
-  
-  KRB_LOCK_MUTEX(text->utils);
-  
-  len=krb_mk_safe(text->enc_in_buf->data, (text->encode_buf+4),
-		  text->enc_in_buf->curlen,
-		  &text->session, &text->ip_local, &text->ip_remote);
-
-  KRB_UNLOCK_MUTEX(text->utils);
-
-  /* returns -1 on error */
-  if (len==-1) return SASL_FAIL;
-
-  /* now copy in the len of the buffer in network byte order */
-  *outputlen=len+4;
-  len=htonl(len);
-  memcpy(text->encode_buf, &len, 4);
-
-  /* Setup the const pointer */
-  *output = text->encode_buf;
-  
-  return SASL_OK;
-}
-
-static int integrity_decode_once(void *context,
-				 const char **input, unsigned *inputlen,
-				 char **output, unsigned *outputlen)
-{
-    int tocopy, result;
-    MSG_DAT data;
-    context_t *text=context;
-    unsigned diff;
-
-    if (text->needsize>0) /* 4 bytes for how long message is */
-    {
-      /* if less than 4 bytes just copy those we have into text->size */
-      if (*inputlen<4) 
-	tocopy=*inputlen;
-      else
-	tocopy=4;
-      
-      if (tocopy>text->needsize)
-	tocopy=text->needsize;
-
-      memcpy(text->sizebuf+4-text->needsize, *input, tocopy);
-      text->needsize-=tocopy;
-
-      *input+=tocopy;
-      *inputlen-=tocopy;
-
-      if (text->needsize==0) /* got all of size */
-      {
-	memcpy(&(text->size), text->sizebuf, 4);
-	text->cursize=0;
-	text->size=ntohl(text->size);
-	if ((text->size>0xFFFF) || (text->size < 0)) return SASL_FAIL; /* too big probably error */
-
-	if (text->bufsize < text->size) {
-	    text->buffer = text->utils->realloc(text->buffer, text->size);
-	    text->bufsize = text->size;
-	}
-	if (text->buffer == NULL) return SASL_NOMEM;
-      }
-
-      *outputlen=0;
-      *output=NULL;
-      if (*inputlen==0) /* have to wait until next time for data */
-	return SASL_OK;
-
-      if (text->size==0)  /* should never happen */
-	return SASL_FAIL;
-    }
-
-    diff=text->size - text->cursize; /* bytes need for full message */
-
-    if (*inputlen < diff) /* not enough for a decode */
-    {
-
-      memcpy(text->buffer+text->cursize, *input, *inputlen);
-      text->cursize+=*inputlen;
-      *inputlen=0;
-      *outputlen=0;
-      *output=NULL;
-      return SASL_OK;
-    } else {
-      memcpy(text->buffer+text->cursize, *input, diff);
-      input+=diff;      
-      *inputlen-=diff;
-    }
-
-    KRB_LOCK_MUTEX(text->utils);
-    
-    result = krb_rd_safe((char *) text->buffer, text->size,
-			 &text->session, &text->ip_remote,
-			 &text->ip_local, &data);
-
-    KRB_UNLOCK_MUTEX(text->utils);
-
-    /* see if the krb library found a problem with what we were given */
-    if (result != 0)
-    {
-	text->utils->seterror(text->utils->conn, 0, krb_err_txt[result]);
-	return SASL_FAIL;
-    }
-
-    /* check to make sure the timestamps are ok */
-    if ((data.time_sec < text->time_sec) || /* if an earlier time */
-	(((data.time_sec == text->time_sec) && /* or the exact same time */
-	 (data.time_5ms < text->time_5ms)))) 
-    {
-	text->utils->seterror(text->utils->conn, 0, "timestamps not ok");
-	return SASL_FAIL;
-    }
-
-    text->time_sec = data.time_sec;
-    text->time_5ms = data.time_5ms;
-
-    *output=text->utils->malloc(data.app_length+1);
-
-    if ((*output) == NULL) return SASL_NOMEM;
- 
-    *outputlen=data.app_length;
-    memcpy((char *)*output, data.app_data,data.app_length);
-
-    text->size=-1;
-    text->needsize=4;
-
-    return SASL_OK;
-}
-
-static int integrity_decode(void *context,
-			    const char *input, unsigned inputlen,
-			    const char **output, unsigned *outputlen)
-{
-    char *tmp = NULL;
-    unsigned tmplen = 0;
-    context_t *text=context;
-    int ret;
-    
-    *outputlen = 0;
-
-    while (inputlen!=0)
-    {
-      ret = integrity_decode_once(text, &input, &inputlen,
-			   &tmp, &tmplen);
-
-      if(ret != SASL_OK) return ret;
-
-      if (tmp!=NULL) /* if received 2 packets merge them together */
-      {
-	  ret = _plug_buf_alloc(text->utils, &text->decode_buf,
-				&text->decode_buf_len, *outputlen + tmplen);
-
-	  *output = text->decode_buf;
-	  memcpy(text->decode_buf + *outputlen, tmp, tmplen);
-	  *outputlen+=tmplen;
-	  text->utils->free(tmp);
-      }
-    }
-
-    return SASL_OK;
-}
-
-
 static int
 new_text(const sasl_utils_t *utils, context_t **text)
 {
     context_t *ret = (context_t *) utils->malloc(sizeof(context_t));
-    if (ret==NULL) return SASL_NOMEM;
+    if (ret==NULL) {
+	MEMERROR(utils);
+	return SASL_NOMEM;
+    }
 
     memset(ret, 0, sizeof(context_t));
 
@@ -593,6 +443,7 @@ server_start(void *glob_context __attribute__((unused)),
 	if(!krb_mutex) {
 	    sparams->utils->seterror(sparams->utils->conn, 0,
 				     "couldn't allocate mutex");
+	    return SASL_FAIL;
 	}
     }
     
@@ -607,6 +458,7 @@ static void dispose(void *conn_context, const sasl_utils_t *utils)
     if (text->buffer) utils->free(text->buffer);
     if (text->encode_buf) utils->free(text->encode_buf);
     if (text->decode_buf) utils->free(text->decode_buf);
+    if (text->decode_once_buf) utils->free(text->decode_once_buf);
     if (text->out_buf) utils->free(text->out_buf);
     if (text->enc_in_buf) {
 	if(text->enc_in_buf->data) utils->free(text->enc_in_buf->data);
@@ -620,7 +472,10 @@ static void dispose(void *conn_context, const sasl_utils_t *utils)
 static void mech_free(void *glob_context __attribute__((unused)),
 		      const sasl_utils_t *utils)
 {
-    if (krb_mutex) utils->mutex_free(krb_mutex);
+    if (krb_mutex) {
+	utils->mutex_free(krb_mutex);
+	krb_mutex = NULL; /* in case we need to re-use it */
+    }
 
     if (srvtab && --refcount == 0) {
 	utils->free(srvtab);
@@ -668,13 +523,11 @@ static int server_continue_step (void *conn_context,
 
   if (text->state==0)
   {    
-      /* random 32-bit number */
-      int randocts, nchal;
+    /* random 32-bit number */
+    int randocts, nchal;
 
-      /* shouldn't we check for erroneous client input here?!? */
+    /* shouldn't we check for erroneous client input here?!? */
 
-    VL(("KERBEROS_V4 Step 1\n"));
-    
     sparams->utils->rand(sparams->utils->rpool,(char *) &randocts ,
 			 sizeof(randocts));    
     text->challenge=randocts; 
@@ -701,17 +554,15 @@ static int server_continue_step (void *conn_context,
     unsigned lup;
     struct sockaddr_in addr;
 
-    VL(("KERBEROS_V4 Step 2\n"));
-
     /* received authenticator */
 
     /* create ticket */
     if (clientinlen > MAX_KTXT_LEN) {
-	text->utils->seterror(text->utils->conn,0,"request larger than maximum ticket size");
+	text->utils->seterror(text->utils->conn,0,
+			      "request larger than maximum ticket size");
 	return SASL_FAIL;
     }
     
-
     ticket.length=clientinlen;
     for (lup=0;lup<clientinlen;lup++)      
       ticket.dat[lup]=clientin[lup];
@@ -734,10 +585,11 @@ static int server_continue_step (void *conn_context,
         from machine to machine) */
 
     /* get ip number in addr*/
-    result = _plug_ipfromstring(sparams->ipremoteport, &addr);
-    if (result != SASL_OK)
-	/* couldn't get remote IP address */
+    result = _plug_ipfromstring(sparams->utils, sparams->ipremoteport, &addr);
+    if (result != SASL_OK || !sparams->ipremoteport) {
+	SETERROR(text->utils, "couldn't get remote IP address");
 	return result;
+    }
 #endif
 
     /* check ticket */
@@ -748,9 +600,9 @@ static int server_continue_step (void *conn_context,
     KRB_UNLOCK_MUTEX(sparams->utils);
 
     if (result) { /* if fails mechanism fails */
-	text->utils->seterror(text->utils->conn,0,krb_err_txt[result]);
-	VL(("krb_rd_req failed service=%s instance=%s error code=%i\n",
-	    sparams->service, text->instance,result));
+	text->utils->seterror(text->utils->conn,0,
+	"krb_rd_req failed service=%s instance=%s error code=%s (%i)",
+	    sparams->service, text->instance,krb_err_txt[result],result);
 	return SASL_BADAUTH;
     }
 
@@ -812,13 +664,18 @@ static int server_continue_step (void *conn_context,
     unsigned char *in;
 
     if ((clientinlen==0) || (clientinlen % 8 != 0)) {
-	text->utils->seterror(text->utils->conn,0,"Response to challengs is not a multiple of 8 octets (a DES block)");
+	text->utils->seterror(text->utils->conn,0,
+			      "Response to challengs is not a multiple of 8 octets (a DES block)");
 	return SASL_FAIL;	
     }
 
     /* we need to make a copy because des does in place decrpytion */
     in = sparams->utils->malloc(clientinlen + 1);
-    if (in == NULL) return SASL_NOMEM;
+    if (in == NULL) {
+	MEMERROR(sparams->utils);
+	return SASL_NOMEM;
+    }
+  
     memcpy(in, clientin, clientinlen);
     in[clientinlen]='\0';
 
@@ -833,46 +690,54 @@ static int server_continue_step (void *conn_context,
 
     testnum=(in[0]*256*256*256)+(in[1]*256*256)+(in[2]*256)+in[3];
 
-    if (testnum != text->challenge)
-	/* incorrect response to challenge */
+    if (testnum != text->challenge) {
+	SETERROR(sparams->utils, "incorrect response to challenge");
 	return SASL_BADAUTH;
-
+    }
+    
     if (! cando_sec(&sparams->props, in[4] & KRB_SECFLAGS)) {
-	/* invalid security property specified */
+	SETERROR(sparams->utils,
+		      "invalid security property specified");
 	return SASL_BADPROT;
     }
+
+    oparams->encode=&encode;
+    oparams->decode=&decode;
     
     switch (in[4] & KRB_SECFLAGS) {
     case KRB_SECFLAG_NONE:
+	text->sec_type = KRB_SEC_NONE;
 	oparams->encode=NULL;
 	oparams->decode=NULL;
 	oparams->mech_ssf=0;
 	break;
     case KRB_SECFLAG_INTEGRITY:
-	oparams->encode=&integrity_encode;
-	oparams->decode=&integrity_decode;
+	text->sec_type = KRB_SEC_INTEGRITY;
 	oparams->mech_ssf=KRB_INTEGRITY_BITS;
 	break;
     case KRB_SECFLAG_ENCRYPTION:
-	oparams->encode=&privacy_encode;
-	oparams->decode=&privacy_decode;
+	text->sec_type = KRB_SEC_ENCRYPTION;
 	oparams->mech_ssf=KRB_DES_SECURITY_BITS;
 	break;
     default:
-        /* not a supported encryption layer */
+        SETERROR(sparams->utils, "not a supported encryption layer");
 	return SASL_BADPROT;
     }
 
     /* get ip data */
     /* get ip number in addr*/
-    result = _plug_ipfromstring(sparams->iplocalport, &(text->ip_local));
+    result = _plug_ipfromstring(sparams->utils,
+				sparams->iplocalport, &(text->ip_local));
     if (result != SASL_OK) {
+        SETERROR(sparams->utils, "couldn't get local ip address");
 	/* couldn't get local IP address */
 	return result;
     }
 
-    result = _plug_ipfromstring(sparams->ipremoteport, &(text->ip_remote));
+    result = _plug_ipfromstring(sparams->utils,
+				sparams->ipremoteport, &(text->ip_remote));
     if (result != SASL_OK) {
+        SETERROR(sparams->utils, "couldn't get remote ip address");
 	/* couldn't get remote IP address */
 	return result;
     }
@@ -897,8 +762,11 @@ static int server_continue_step (void *conn_context,
       }
 
       authid = sparams->utils->malloc(alen + 1);
-      if (!authid)
-	return SASL_NOMEM;
+      if (!authid) {
+	  MEMERROR(sparams->utils);
+	  return SASL_NOMEM;
+      }
+      
       strcpy(authid, text->pname);
       if (text->pinst[0]) {
 	strcat(authid, ".");
@@ -911,7 +779,11 @@ static int server_continue_step (void *conn_context,
 
       if (in[8]) {
 	  user = sparams->utils->malloc(strlen((char *) in + 8) + 1);
-	  if (!user) return SASL_NOMEM;
+	  if (!user) {
+	      MEMERROR(sparams->utils);
+	      return SASL_NOMEM;
+	  }
+	  
 	  strcpy(user, (char *) in + 8);
 	  ulen = strlen(user);
       } else {
@@ -945,6 +817,7 @@ static int server_continue_step (void *conn_context,
   }
   
   /* Improper step. Probably application error */
+  SETERROR(sparams->utils, "invalid step");
   return SASL_FAIL; /* should never get here */
 }
 
@@ -969,12 +842,11 @@ static sasl_server_plug_t plugins[] =
 };
 #endif
 
-int sasl_server_plug_init(const sasl_utils_t *utils,
-			  int maxversion,
-			  int *out_version,
-			  sasl_server_plug_t **pluglist,
-			  int *plugcount,
-			  const char *plugname __attribute__((unused)))
+int kerberos4_server_plug_init(const sasl_utils_t *utils,
+			       int maxversion,
+			       int *out_version,
+			       sasl_server_plug_t **pluglist,
+			       int *plugcount)
 {
 #ifdef macintosh
 	return SASL_BADVERS;
@@ -987,7 +859,8 @@ int sasl_server_plug_init(const sasl_utils_t *utils,
     }
 
     if(!srvtab) {	
-	utils->getopt(utils->getopt_context, "KERBEROS_V4", "srvtab", &ret, &rl);
+	utils->getopt(utils->getopt_context,
+		      "KERBEROS_V4", "srvtab", &ret, &rl);
 
 	if (ret == NULL) {
 	    ret = KEYFILE;
@@ -995,7 +868,10 @@ int sasl_server_plug_init(const sasl_utils_t *utils,
 	}
     
 	srvtab = utils->malloc(sizeof(char) * (rl + 1));
-	if(!srvtab) return SASL_NOMEM;
+	if(!srvtab) {
+	    MEMERROR(utils);
+	    return SASL_NOMEM;
+	}
 
 	strcpy(srvtab, ret);
     }
@@ -1023,8 +899,6 @@ static int client_start(void *glob_context __attribute__((unused)),
 		 sasl_client_params_t *params,
 		 void **conn)
 {
-  VL(("KERBEROS_V4 Client start\n"));
-
   return new_text(params->utils, (context_t **) conn);
 }
 
@@ -1043,18 +917,17 @@ static int client_continue_step (void *conn_context,
 
     authent.length = MAX_KTXT_LEN;
   
-    if (text->state==0)
+    if (text->state == 0 && serverinlen == 0)
     {
-	VL(("KERBEROS_V4 Step 1\n"));
-
-	if (clientout) {
-	    *clientout = NULL;
-	    *clientoutlen = 0;
-	}
+	/* KERBEROS_V4 can't do client-first */
 
+	*clientout = NULL;
+	*clientoutlen = 0;
 	text->state=1;
 
 	return SASL_CONTINUE;
+    } else if(text->state == 0) {
+	text->state = 1;
     }
 
     else if (text->state==1) {
@@ -1063,17 +936,13 @@ static int client_continue_step (void *conn_context,
 	int result;
 	KTEXT_ST ticket;
 
-	VL(("KERBEROS_V4 Step 2\n"));
-
 	memset(&ticket, 0L, sizeof(ticket));
 	ticket.length=MAX_KTXT_LEN;   
 
 	if (serverinlen != 4) {
-	    text->utils->seterror(text->utils->conn, SASL_NOLOG, "server challenge not 4 bytes long");
-	    
-	    cparams->utils->log(NULL, SASL_LOG_ERR,
-			       "server challenge not 4 bytes long");
-	    return SASL_FAIL; 
+	    text->utils->seterror(text->utils->conn, 0,
+				  "server challenge not 4 bytes long");
+	    return SASL_BADPROT; 
 	}
 
 	memcpy(&text->challenge, serverin, 4);
@@ -1083,11 +952,13 @@ static int client_continue_step (void *conn_context,
 	if (cparams->serverFQDN == NULL) {
 	    cparams->utils->log(NULL, SASL_LOG_ERR,
 				"no 'serverFQDN' set");
+	    SETERROR(text->utils, "paramater error");
 	    return SASL_BADPARAM;
 	}
 	if (cparams->service == NULL) {
 	    cparams->utils->log(NULL, SASL_LOG_ERR,
 				"no 'service' set");
+	    SETERROR(text->utils, "paramater error");
 	    return SASL_BADPARAM;
 	}
 
@@ -1121,8 +992,7 @@ static int client_continue_step (void *conn_context,
 	    KRB_UNLOCK_MUTEX(cparams->utils);
 	    
 	    text->utils->seterror(text->utils->conn,SASL_NOLOG,
-			  "krb_mk_req() failed: %s (%d)",
-			  krb_err_txt[result], result);
+			  "krb_mk_req() failed");
 	    
 	    cparams->utils->log(NULL, SASL_LOG_ERR, 
 			       "krb_mk_req() failed: %s (%d)",
@@ -1169,8 +1039,10 @@ static int client_continue_step (void *conn_context,
 	    for (prompt = *prompt_need;
 		 prompt->id != SASL_CB_LIST_END;
 		 ++prompt)
-		if (! prompt->result)
+		if (! prompt->result) {
+		    PARAMERROR(cparams->utils);
 		    return SASL_BADPARAM;
+		}
 
 	    /* Get the username */
 	    if (! text->user) {
@@ -1179,7 +1051,11 @@ static int client_continue_step (void *conn_context,
 		     ++prompt)
 		    if (prompt->id == SASL_CB_USER) {
 			text->user = cparams->utils->malloc(strlen(prompt->result) + 1);
-			if(!text->user) return SASL_NOMEM;
+			if(!text->user) {
+			    MEMERROR(cparams->utils);
+			    return SASL_NOMEM;
+			}
+			
 			strcpy(text->user, prompt->result);
 			
 			break;
@@ -1211,15 +1087,22 @@ static int client_continue_step (void *conn_context,
 				    SASL_CB_USER,
 				    &userid,
 				    NULL);
-		if (result != SASL_OK)
+		if (result != SASL_OK) {
+		    SETERROR(cparams->utils, "getuser callback failed");
 		    return result;
+		}
+		
 		if (userid) {		    
 		    text->user = cparams->utils->malloc(strlen(userid) + 1);
-		    if (!text->user) return SASL_NOMEM;
+		    if (!text->user) {
+			MEMERROR(cparams->utils);
+			return SASL_NOMEM;
+		    }
 		    strcpy(text->user, userid);
 		}
 		break;
 	    default:
+		SETERROR(cparams->utils, "couldn't get callback");
 		return result;
 	    }
 	}
@@ -1230,8 +1113,10 @@ static int client_continue_step (void *conn_context,
 	    if (! prompt_need)
 		return SASL_INTERACT;
 	    *prompt_need = cparams->utils->malloc(sizeof(sasl_interact_t) * 2);
-	    if (! *prompt_need)
+	    if (! *prompt_need) {
+		MEMERROR(cparams->utils);
 		return SASL_NOMEM;
+	    }
 	    prompt = *prompt_need;
 	    prompt->id = SASL_CB_USER;
 	    prompt->prompt = "Remote Userid";
@@ -1243,8 +1128,8 @@ static int client_continue_step (void *conn_context,
       
 	/* must be 8 octets */
 	if (serverinlen!=8) {
-	    cparams->utils->log(NULL, SASL_LOG_ERR,
-			        "server response not 8 bytes long");
+	    SETERROR(cparams->utils,
+		     "server response not 8 bytes long");
 	    return SASL_BADAUTH;
 	}
 
@@ -1263,6 +1148,7 @@ static int client_continue_step (void *conn_context,
 	    cparams->utils->log(NULL, SASL_LOG_ERR,
 				"krb_get_cred() failed: %s (%d)",
 				krb_err_txt[result], result);
+	    SETERROR(cparams->utils, "krb_get_cred() failed");
 	    return SASL_BADAUTH;
 	}
 #endif
@@ -1283,8 +1169,7 @@ static int client_continue_step (void *conn_context,
 	/* verify data 1st 4 octets must be equal to chal+1 */
 	if (testnum != text->challenge+1)
 	{
-	    cparams->utils->log(NULL, SASL_LOG_ERR,
-				"server response incorrect");
+	    SETERROR(cparams->utils,"server response incorrect");
 	    return SASL_BADAUTH;
 	}
 
@@ -1295,17 +1180,21 @@ static int client_continue_step (void *conn_context,
 	 */
 	if (cparams->props.min_ssf > 
 	       KRB_DES_SECURITY_BITS + cparams->external_ssf) {
-	    cparams->utils->log(NULL, SASL_LOG_ERR,
+	    SETERROR(cparams->utils,
 			       "minimum ssf too strong for this mechanism");
 	    return SASL_TOOWEAK;
 	} else if (cparams->props.min_ssf > cparams->props.max_ssf) {
-	    cparams->utils->log(NULL, SASL_LOG_ERR,
+	    SETERROR(cparams->utils,
 			       "minimum ssf larger than maximum ssf");
 	    return SASL_BADPARAM;
 	}
 
 	/* create stuff to send to server */
 	sout = (char *) cparams->utils->malloc(9+strlen(text->user)+9);
+	if(!sout) {
+	    MEMERROR(cparams->utils);
+	    return SASL_NOMEM;
+	}
 
 	nchal=htonl(text->challenge);
 	memcpy(sout, &nchal, 4);
@@ -1314,30 +1203,32 @@ static int client_continue_step (void *conn_context,
 	need = cparams->props.max_ssf - cparams->external_ssf;
 	musthave = cparams->props.min_ssf - cparams->external_ssf;
 
+	oparams->decode = &decode;
+	oparams->encode = &encode;
+
 	if ((in[4] & KRB_SECFLAG_ENCRYPTION)
 	    && (need>=56) && (musthave <= 56)) {
 	    /* encryption */
-	    oparams->encode = &privacy_encode;
-	    oparams->decode = &privacy_decode;
+	    text->sec_type = KRB_SEC_ENCRYPTION;
 	    oparams->mech_ssf = 56;
 	    sout[4] = KRB_SECFLAG_ENCRYPTION;
 	    /* using encryption layer */
 	} else if ((in[4] & KRB_SECFLAG_INTEGRITY)
 		   && (need >= 1) && (musthave <= 1)) {
 	    /* integrity */
-	    oparams->encode=&integrity_encode;
-	    oparams->decode=&integrity_decode;
+	    text->sec_type = KRB_SEC_INTEGRITY;
 	    oparams->mech_ssf=1;
 	    sout[4] = KRB_SECFLAG_INTEGRITY;
 	    /* using integrity layer */
 	} else if ((in[4] & KRB_SECFLAG_NONE) && (musthave <= 0)) {
 	    /* no layer */
+	    text->sec_type = KRB_SEC_NONE;
 	    oparams->encode=NULL;
 	    oparams->decode=NULL;
 	    oparams->mech_ssf=0;
 	    sout[4] = KRB_SECFLAG_NONE;
 	} else {
-	    cparams->utils->log(NULL, SASL_LOG_ERR,
+	    SETERROR(cparams->utils,
 				"unable to agree on layers with server");
 	    return SASL_BADPROT;
 	}
@@ -1373,7 +1264,10 @@ static int client_continue_step (void *conn_context,
 			 (des_cblock *)text->session,
 			 DES_ENCRYPT);
 
-	_plug_buf_alloc(text->utils, &text->out_buf, &text->out_buf_len, len);
+	result = _plug_buf_alloc(text->utils, &text->out_buf,
+				 &text->out_buf_len, len);
+	if(result != SASL_OK)
+	    return result;
 
 	memcpy(text->out_buf, sout, len);
 
@@ -1382,7 +1276,8 @@ static int client_continue_step (void *conn_context,
 
 	/* nothing more to do; should be authenticated */
 	if(cparams->iplocalport) {   
-	    result = _plug_ipfromstring(cparams->iplocalport,
+	    result = _plug_ipfromstring(cparams->utils,
+					cparams->iplocalport,
 					&(text->ip_local));
 	    if (result != SASL_OK) {
 		/* couldn't get local IP address */
@@ -1391,7 +1286,8 @@ static int client_continue_step (void *conn_context,
 	}
 	
 	if(cparams->ipremoteport) {
-	    result = _plug_ipfromstring(cparams->ipremoteport,
+	    result = _plug_ipfromstring(cparams->utils,
+					cparams->ipremoteport,
 					&(text->ip_remote));
 	    if (result != SASL_OK) {
 		/* couldn't get local IP address */
@@ -1402,8 +1298,10 @@ static int client_continue_step (void *conn_context,
 	buf = cparams->utils->malloc(strlen(text->credentials.pname)
 				   + strlen(text->credentials.pinst)
 				   + 2);
-	if (!buf)
+	if (!buf) {
+	    MEMERROR(cparams->utils);
 	    return SASL_NOMEM;
+	}
 	strcpy(buf, text->credentials.pname);
 	if (text->credentials.pinst[0]) {
 	    strcat(buf, ".");
@@ -1446,6 +1344,7 @@ static int client_continue_step (void *conn_context,
 	return SASL_OK;
     }
 
+    SETERROR(cparams->utils, "shouldn't ever get here");
     return SASL_FAIL; /* should never get here */
 }
 
@@ -1473,23 +1372,25 @@ static sasl_client_plug_t client_plugins[] =
   }
 };
 
-int sasl_client_plug_init(const sasl_utils_t *utils __attribute__((unused)),
-			  int maxversion,
-			  int *out_version,
-			  sasl_client_plug_t **pluglist,
-			  int *plugcount,
-			  const char *plugname __attribute__((unused)))
+int kerberos4_client_plug_init(
+    const sasl_utils_t *utils,
+    int maxversion,
+    int *out_version,
+    sasl_client_plug_t **pluglist,
+    int *plugcount)
 {
-  if (maxversion < SASL_CLIENT_PLUG_VERSION)
-    return SASL_BADVERS;
+    if (maxversion < SASL_CLIENT_PLUG_VERSION) {
+	SETERROR(utils, "Wrong KERBEROS_V4 version");
+	return SASL_BADVERS;
+    }
 
-  *pluglist=client_plugins;
+    *pluglist=client_plugins;
 
-  *plugcount=1;
-  *out_version=SASL_CLIENT_PLUG_VERSION;
+    *plugcount=1;
+    *out_version=SASL_CLIENT_PLUG_VERSION;
 
-  refcount++;
+    refcount++;
 
-  return SASL_OK;
+    return SASL_OK;
 }
 
