@@ -1,7 +1,7 @@
 /* Plain SASL plugin
  * Rob Siemborski
  * Tim Martin 
- * $Id: plain.c,v 1.43.2.7 2001/06/27 14:56:31 rjs3 Exp $
+ * $Id: plain.c,v 1.43.2.13 2001/07/12 14:10:14 rjs3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -82,7 +82,10 @@ static int start(void *glob_context __attribute__((unused)),
 
   /* holds state are in */
   text=sparams->utils->malloc(sizeof(context_t));
-  if (text==NULL) return SASL_NOMEM;
+  if (text==NULL) {
+      MEMERROR(sparams->utils);
+      return SASL_NOMEM;
+  }
 
   memset(text, 0, sizeof(context_t));
 
@@ -174,8 +177,6 @@ server_continue_step (void *conn_context,
 
     /* should have received author-id NUL authen-id NUL password */
 
-    VL (("in=%s len=%i\n",clientin,clientinlen));
-
     /* get author */
     author = clientin;
     while ((lup < clientinlen) && (clientin[lup] != 0))
@@ -183,7 +184,7 @@ server_continue_step (void *conn_context,
 
     if (lup >= clientinlen)
     {
-	VL(("Can only find author (no password)\n"));
+	SETERROR(params->utils, "Can only find author (no password)");
 	return SASL_BADPROT;
     }
 
@@ -195,7 +196,8 @@ server_continue_step (void *conn_context,
 
     if (lup >= clientinlen)
     {
-	VL(("Can only find author/en (no password) \n"));
+	params->utils->seterror(params->utils->conn, 0,
+				"Can only find author/en (no password)");
 	return SASL_BADPROT;
     }
 
@@ -208,6 +210,7 @@ server_continue_step (void *conn_context,
     password_len = clientin + lup - password;
 
     if (lup != clientinlen) {
+	SETERROR(params->utils, "Got more data than we were expecting in the PLAIN plugin\n");
 	return SASL_BADPROT;
     }
     
@@ -215,7 +218,10 @@ server_continue_step (void *conn_context,
        but we can't assume there is an allocated byte at the end
        of password so we have to copy it */
     passcopy = params->utils->malloc(password_len + 1);    
-    if (passcopy == NULL) return SASL_NOMEM;
+    if (passcopy == NULL) {
+	MEMERROR(params->utils);
+	return SASL_NOMEM;
+    }
 
     strncpy(passcopy, password, password_len);
     passcopy[password_len] = '\0';
@@ -227,12 +233,11 @@ server_continue_step (void *conn_context,
 
     if (result != SASL_OK)
     {
-	VL(("Password verification failed\n"));
+	params->utils->seterror(params->utils->conn, 0,
+				"Password verification failed");
 	return result;
     }
 
-    VL(("Password verified A-ok\n"));
-
     if (! author || !*author)
       author = authen;
 
@@ -256,6 +261,8 @@ server_continue_step (void *conn_context,
     return SASL_OK;
   }
 
+  SETERROR( params->utils,
+	    "Unexpected State Reached in PLAIN plugin");
   return SASL_FAIL; /* should never get here */
 }
 
@@ -279,22 +286,23 @@ static sasl_server_plug_t plugins[] =
   }
 };
 
-int sasl_server_plug_init(const sasl_utils_t *utils __attribute__((unused)),
-			  int maxversion,
-			  int *out_version,
-			  sasl_server_plug_t **pluglist,
-			  int *plugcount,
-			  const char *plugname __attribute__((unused)))
+int plain_server_plug_init(const sasl_utils_t *utils,
+			   int maxversion,
+			   int *out_version,
+			   sasl_server_plug_t **pluglist,
+			   int *plugcount)
 {
-  if (maxversion<SASL_SERVER_PLUG_VERSION)
-    return SASL_BADVERS;
+    if (maxversion<SASL_SERVER_PLUG_VERSION) {
+	SETERROR(utils, "PLAIN version mismatch");
+	return SASL_BADVERS;
+    }
+    
+    *pluglist=plugins;
 
-  *pluglist=plugins;
+    *plugcount=1;  
+    *out_version=SASL_SERVER_PLUG_VERSION;
 
-  *plugcount=1;  
-  *out_version=SASL_SERVER_PLUG_VERSION;
-
-  return SASL_OK;
+    return SASL_OK;
 }
 
 /* put in sasl_wrongmech */
@@ -302,26 +310,27 @@ static int c_start(void *glob_context __attribute__((unused)),
 		   sasl_client_params_t *params,
 		   void **conn)
 {
-  context_t *text;
+    context_t *text;
 
-  /* holds state are in */
-  text = params->utils->malloc(sizeof(context_t));
-  if (text==NULL) return SASL_NOMEM;
+    /* holds state are in */
+    text = params->utils->malloc(sizeof(context_t));
+    if (text==NULL) {
+	MEMERROR( params->utils );
+	return SASL_NOMEM;
+    }
+    
+    memset(text, 0, sizeof(context_t));
 
-  memset(text, 0, sizeof(context_t));
+    text->state=1;
+    *conn=text;
 
-  text->state=1;
-
-  *conn=text;
-
-  return SASL_OK;
+    return SASL_OK;
 }
 
 /* 
  * Trys to find the prompt with the lookingfor id in the prompt list
  * Returns it if found. NULL otherwise
  */
-
 static sasl_interact_t *find_prompt(sasl_interact_t **promptlist,
 				    unsigned int lookingfor)
 {
@@ -372,8 +381,11 @@ static int get_userid(sasl_client_params_t *params,
 			NULL);
     if (result != SASL_OK)
       return result;
-    if (! id)
-      return SASL_BADPARAM;
+    if (! id) {
+	PARAMERROR(params->utils);
+	return SASL_BADPARAM;
+    }
+    
     *userid = id;
   }
 
@@ -413,8 +425,11 @@ static int get_authid(sasl_client_params_t *params,
 			NULL);
     if (result != SASL_OK)
       return result;
-    if (! id)
-      return SASL_BADPARAM;
+    if (! id) {
+	PARAMERROR( params->utils );
+	return SASL_BADPARAM;
+    }
+    
     *authid = id;
   }
 
@@ -435,21 +450,26 @@ static int get_password(sasl_client_params_t *params,
   prompt=find_prompt(prompt_need,SASL_CB_PASS);
   if (prompt!=NULL)
   {
-    /* We prompted, and got.*/
+      /* We prompted, and got.*/
 	
-    if (! prompt->result)
-      return SASL_FAIL;
+      if (! prompt->result) {
+	  SETERROR(params->utils, "Unexpectedly missing a prompt result");
+	  return SASL_FAIL;
+      }
+      
+      /* copy what we got into a secret_t */
+      *password = (sasl_secret_t *) params->utils->malloc(sizeof(sasl_secret_t)+
+							  prompt->len+1);
+      if (! *password) {
+	  MEMERROR( params->utils );
+	  return SASL_NOMEM;
+      }
+      
+      (*password)->len=prompt->len;
+      memcpy((*password)->data, prompt->result, prompt->len);
+      (*password)->data[(*password)->len]=0;
 
-    /* copy what we got into a secret_t */
-    *password = (sasl_secret_t *) params->utils->malloc(sizeof(sasl_secret_t)+
-						       prompt->len+1);
-    if (! *password) return SASL_NOMEM;
-
-    (*password)->len=prompt->len;
-    memcpy((*password)->data, prompt->result, prompt->len);
-    (*password)->data[(*password)->len]=0;
-
-    return SASL_OK;
+      return SASL_OK;
   }
 
 
@@ -484,10 +504,17 @@ static int make_prompts(sasl_client_params_t *params,
   if (auth_res==SASL_INTERACT) num++;
   if (pass_res==SASL_INTERACT) num++;
 
-  if (num==1) return SASL_FAIL;
+  if (num==1) {
+      SETERROR( params->utils, "make_prompts called with no actual prompts" );
+      return SASL_FAIL;
+  }
 
   prompts=params->utils->malloc(sizeof(sasl_interact_t)*(num+1));
-  if ((prompts) ==NULL) return SASL_NOMEM;
+  if ((prompts) ==NULL) {
+      MEMERROR( params->utils );
+      return SASL_NOMEM;
+  }
+  
   *prompts_res=prompts;
 
   if (user_res==SASL_INTERACT)
@@ -538,7 +565,7 @@ static int make_prompts(sasl_client_params_t *params,
 static int client_continue_step (void *conn_context,
 				 sasl_client_params_t *params,
 				 const char *serverin __attribute__((unused)),
-				 unsigned serverinlen,
+				 unsigned serverinlen __attribute__((unused)),
 				 sasl_interact_t **prompt_need,
 				 const char **clientout,
 				 unsigned *clientoutlen,
@@ -550,30 +577,25 @@ static int client_continue_step (void *conn_context,
   context_t *text;
   text=conn_context;
 
-  VL(("Plain step #%d\n",text->state));
-
-  if (!clientout && text->state == 1) {
-      text->state = 2;
-      return SASL_CONTINUE;
-  }
   *clientout = NULL;
   *clientoutlen = 0;
 
   /* doesn't really matter how the server responds */
 
-  if (text->state==1 || (text->state == 2 && serverinlen == 0)) {
+  if (text->state==1) {
     int user_result=SASL_OK;
     int auth_result=SASL_OK;
     int pass_result=SASL_OK;
 
     /* check if sec layer strong enough */
-    if (params->props.min_ssf>0)
-      return SASL_TOOWEAK;
+    if (params->props.min_ssf>0+params->external_ssf) {
+	SETERROR( params->utils, "The PLAIN plugin cannot support any SSF");
+	return SASL_TOOWEAK;
+    }
 
     /* try to get the userid */
     if (oparams->user==NULL)
     {
-      VL (("Trying to get userid\n"));
       user_result=get_userid(params,
 			     &user,
 			     prompt_need);
@@ -585,7 +607,6 @@ static int client_continue_step (void *conn_context,
     /* try to get the authid */    
     if (oparams->authid==NULL)
     {
-      VL (("Trying to get authid\n"));
       auth_result=get_authid(params,
 			     &authid,
 			     prompt_need);
@@ -597,7 +618,6 @@ static int client_continue_step (void *conn_context,
     /* try to get the password */
     if (text->password==NULL)
     {
-      VL (("Trying to get password\n"));
       pass_result=get_password(params,
 			       &text->password,
 			       prompt_need);
@@ -607,7 +627,7 @@ static int client_continue_step (void *conn_context,
     }
 
     /* free prompts we got */
-    if (prompt_need) {
+    if (prompt_need && *prompt_need) {
 	params->utils->free(*prompt_need);
 	*prompt_need = NULL;
     }
@@ -621,17 +641,16 @@ static int client_continue_step (void *conn_context,
 			  user_result, auth_result, pass_result);
       if (result!=SASL_OK) return result;
       
-      VL(("returning prompt(s)\n"));
       return SASL_INTERACT;
     }
     
     params->canon_user(params->utils->conn, user, 0, authid, 0, 0, oparams);
 
-    if (!oparams->authid || !text->password)
-      return SASL_BADPARAM;
-
-    VL (("Got username, authid, and password\n"));
-
+    if (!oparams->authid || !text->password) {
+	PARAMERROR(params->utils);
+	return SASL_BADPARAM;
+    }
+    
     /* send authorized id NUL authentication id NUL password */
     {
       *clientoutlen = (oparams->ulen + 1
@@ -644,11 +663,6 @@ static int client_continue_step (void *conn_context,
       if(result != SASL_OK) return result;
 
       memset(text->out_buf, 0, *clientoutlen + 1);
-
-      VL(("userid=[%s]\n",oparams->user ? oparams->user : "(NULL)"));
-      VL(("authid=[%s]\n",oparams->authid));
-      VL(("password=[%s]\n",text->password->data));
-
       memcpy(text->out_buf, oparams->user, oparams->ulen);
       memcpy(text->out_buf+oparams->ulen+1, oparams->authid, oparams->alen);
       memcpy(text->out_buf+oparams->ulen+oparams->alen+2,
@@ -666,16 +680,15 @@ static int client_continue_step (void *conn_context,
 
     oparams->param_version = 0;
 
-    text->state = 3;
+    text->state = 2;
 
     return SASL_CONTINUE;
   }
 
-  if (text->state == 3)
+  if (text->state == 2)
   {
       *clientout = NULL;
       *clientoutlen = 0;
-      VL(("Verify we're done step"));
       text->state++;
       return SASL_OK;      
   }
@@ -702,20 +715,21 @@ static sasl_client_plug_t client_plugins[] =
   }
 };
 
-int sasl_client_plug_init(sasl_utils_t *utils __attribute__((unused)),
-			  int maxversion,
-			  int *out_version,
-			  sasl_client_plug_t **pluglist,
-			  int *plugcount,
-			  const char *plugname __attribute__((unused)))
+int plain_client_plug_init(sasl_utils_t *utils,
+			   int maxversion,
+			   int *out_version,
+			   sasl_client_plug_t **pluglist,
+			   int *plugcount)
 {
-  if (maxversion<SASL_CLIENT_PLUG_VERSION)
-    return SASL_BADVERS;
+    if (maxversion<SASL_CLIENT_PLUG_VERSION) {
+	SETERROR(utils, "PLAIN version mismatch");
+	return SASL_BADVERS;
+    }
 
-  *pluglist=client_plugins;
+    *pluglist=client_plugins;
 
-  *plugcount=1;
-  *out_version=SASL_CLIENT_PLUG_VERSION;
+    *plugcount=1;
+    *out_version=SASL_CLIENT_PLUG_VERSION;
 
-  return SASL_OK;
+    return SASL_OK;
 }
