@@ -1,7 +1,7 @@
 /* testsuite.c -- Stress the library a little
  * Rob Siemborski
  * Tim Martin
- * $Id: testsuite.c,v 1.13.2.16 2001/06/28 19:35:44 rjs3 Exp $
+ * $Id: testsuite.c,v 1.13.2.26 2001/07/13 17:31:15 rjs3 Exp $
  */
 /* 
  * Copyright (c) 2001 Carnegie Mellon University.  All rights reserved.
@@ -65,7 +65,7 @@
 
 #include <sasl/sasl.h>
 #include <sasl/saslutil.h>
-
+#include <sasl/prop.h>
 #include <sasl/md5global.h>
 #include <sasl/md5.h>
 #include <sasl/hmac-md5.h>
@@ -95,6 +95,8 @@ const char *username = "rjs3";
 const char *nonexistant_username = "ABCDEFGHIJ";
 const char *authname = "rjs3";
 const char *password = "1234";
+const char *cu_plugin = "INTERNAL";
+char other_result[1024];
 
 static const char *gssapi_service = "host";
 
@@ -146,6 +148,8 @@ int DETAILED_MEMORY_DEBUGGING = 0;
 
 mem_info_t *head = NULL;
 
+#ifndef WITH_DMALLOC
+
 void *test_malloc(size_t size)
 {
     void *out;
@@ -262,8 +266,11 @@ void test_free(void *ptr)
     free(ptr);
 }
 
+#endif WITH_DMALLOC
+
 int mem_stat() 
 {
+#ifndef WITH_DMALLOC
     mem_info_t *cur;
 
     if(!head) {
@@ -276,7 +283,12 @@ int mem_stat()
 	fprintf(stderr, "    %X\n", (unsigned)cur->addr);
     }
     return SASL_FAIL;
+#else
+    return SASL_OK;
+#endif /* WITH_DMALLOC */
 }
+
+
 /************* End Memory Allocation functions ******/
 
 void fatal(char *str)
@@ -355,6 +367,11 @@ int good_getopt(void *context __attribute__((unused)),
 	if (len)
 	    *len = strlen("./sasldb");
 	return SASL_OK;
+    } else if (!strcmp(option, "canon_user_plugin")) {
+	*result = cu_plugin;
+	if (len)
+	    *len = strlen(*result);
+	return SASL_OK;
     }
 
     return SASL_FAIL;
@@ -426,10 +443,13 @@ void init(unsigned int seed)
 		   (sasl_mutex_unlock_t *) &my_mutex_unlock,
 		   (sasl_mutex_free_t *) &my_mutex_dispose);
 
+#ifndef WITH_DMALLOC
     sasl_set_alloc((sasl_malloc_t *)test_malloc,
 		   (sasl_calloc_t *)test_calloc,
 		   (sasl_realloc_t *)test_realloc,
 		   (sasl_free_t *)test_free);
+#endif
+
 }
 
 /*
@@ -498,12 +518,13 @@ void test_listmech(void)
 
 
 
-    if (sasl_server_init(emptysasl_cb,"TestSuite")!=SASL_OK) fatal("");
+    if (sasl_server_init(emptysasl_cb,"TestSuite")!=SASL_OK)
+	fatal("can't sasl_server_init");
 
     if (sasl_server_new("rcmd", myhostname,
 			NULL, NULL, NULL, NULL, 0, 
 			&saslconn) != SASL_OK) {
-	fatal("");
+	fatal("can't sasl_server_new");
     }
 
 
@@ -699,6 +720,98 @@ void test_64(void)
     
 }
 
+/* This isn't complete, but then, what in the testsuite is? */
+void test_props(void) 
+{
+    int result;
+    struct propval foobar[3];
+    struct propctx *ctx, *dupctx;
+
+    const char *requests[] = {
+	"userPassword",
+        "userName",
+	"homeDirectory",
+        "uidNumber",
+        "gidNumber",
+        NULL
+    };
+
+    const char *short_requests[] = {
+	"userPassword",
+	"userName",
+	"BAD",
+	NULL
+    };
+
+    ctx = prop_new(2);
+    if(!ctx) {
+	fatal("no new prop context");
+    }
+
+    if(prop_request(NULL, requests) == SASL_OK)
+	fatal("prop_request w/NULL context succeeded");
+    if(prop_request(ctx, NULL) == SASL_OK)
+	fatal("prop_request w/NULL request list succeeded");
+    
+    result = prop_request(ctx, requests);
+    if(result != SASL_OK)
+	fatal("prop request failed");
+
+    prop_set(ctx, "userPassword", "pw1", 0);
+    prop_set(ctx, "userPassword", "pw2", 0);
+    prop_set(ctx, "userName", "rjs3", 0);
+    prop_set(ctx, NULL, "tmartin", 0);
+    
+    if(prop_set(ctx, "gah", "ack", 0) == SASL_OK) {
+	printf("setting bad property name succeeded\n");
+	exit(1);
+    }
+
+    result = prop_getnames(ctx, short_requests, foobar);
+    if(result < 0)
+	fatal("prop_getnames failed");
+
+    if(strcmp(foobar[0].name, short_requests[0]))
+	fatal("prop_getnames item 0 wrong name");
+    if(strcmp(foobar[1].name, short_requests[1]))
+	fatal("prop_getnames item 1 wrong name");
+    if(foobar[2].name)
+	fatal("prop_getnames returned an item 2");
+
+    result = prop_dup(ctx, &dupctx);
+    if(result != SASL_OK)
+	fatal("could not duplicate");
+
+    prop_clear(ctx, 1);
+    
+    result = prop_getnames(ctx, short_requests, foobar);
+    if(result < 0)
+	fatal("prop_getnames failed second time");
+
+    if(foobar[0].name)
+	fatal("it appears that prop_clear failed");
+    
+    result = prop_getnames(dupctx, short_requests, foobar);
+    if(result < 0)
+	fatal("prop_getnames failed second time");
+
+    if(!foobar[0].name)
+	fatal("prop_clear appears to have affected dup'd context");
+    
+    prop_clear(dupctx, 0);
+
+    result = prop_getnames(dupctx, short_requests, foobar);
+    if(result < 0)
+	fatal("prop_getnames failed second time");
+
+    if(!foobar[0].name || strcmp(foobar[0].name, short_requests[0]))
+	fatal("prop_clear appears to have cleared too much");
+
+    prop_dispose(&ctx);
+    prop_dispose(&dupctx);
+    if(ctx != NULL)
+	fatal("ctx not null after prop_dispose");
+}
 
 /* callbacks we support */
 static sasl_callback_t client_callbacks[] = {
@@ -718,29 +831,26 @@ static sasl_callback_t client_callbacks[] = {
 };
 
 void interaction (int id, const char *prompt,
-		  char **tresult, unsigned int *tlen)
+		  const char **tresult, unsigned int *tlen)
 {
-    /* FIXME: Look at that memory leak! */
-    
     if (id==SASL_CB_PASS) {
-	*tresult=(char *) strdup(password);
+	*tresult=(char *) password;
     } else if (id==SASL_CB_USER) {
-	*tresult=(char *) strdup(username);
+	*tresult=(char *) username;
     } else if (id==SASL_CB_AUTHNAME) {
-	*tresult=(char *) strdup(authname);
+	*tresult=(char *) authname;
 #ifdef SASL_CB_GETREALM
     } else if ((id==SASL_CB_GETREALM)) {
-	*tresult=(char *) strdup(myhostname);
+	*tresult=(char *) myhostname;
 #endif
     } else {
-	char result[1024];
 	int c;
 	
 	printf("%s: ",prompt);
-	fgets(result, sizeof(result) - 1, stdin);
-	c = strlen(result);
-	result[c - 1] = '\0';
-	*tresult=strdup(result);
+	fgets(other_result, sizeof(other_result) - 1, stdin);
+	c = strlen(other_result);
+	other_result[c - 1] = '\0';
+	*tresult=other_result;
     }
 
     *tlen=strlen(*tresult);
@@ -767,17 +877,22 @@ const sasl_security_properties_t security_props = {
     NULL	    
 };
 
-void set_properties(sasl_conn_t *conn)
+void set_properties(sasl_conn_t *conn, const sasl_security_properties_t *props)
 {
-  if (sasl_setprop(conn, SASL_SEC_PROPS, &security_props)!=SASL_OK)
-      fatal("sasl_setprop() failed");
+    if(!props) {
+	if (sasl_setprop(conn, SASL_SEC_PROPS, &security_props)!=SASL_OK)
+	    fatal("sasl_setprop() failed");
+    } else {
+       	if (sasl_setprop(conn, SASL_SEC_PROPS, props)!=SASL_OK)
+	    fatal("sasl_setprop() failed");
+    }
 }
 
 /*
  * This corrupts the string for us
  */
-
-void corrupt(corrupt_type_t type, char *in, int inlen, char **out, unsigned *outlen)
+void corrupt(corrupt_type_t type, char *in, int inlen,
+	     char **out, unsigned *outlen)
 {
     unsigned lup;
     
@@ -830,7 +945,9 @@ void corrupt(corrupt_type_t type, char *in, int inlen, char **out, unsigned *out
 	case SHORTEN:
 	    if (inlen > 0)
 	    {
-		*outlen = (rand() % inlen);
+		*outlen = 0;
+		while(*outlen == 0)
+		    *outlen = (rand() % inlen);
 		*out = (char *) malloc(*outlen);
 		memcpy(*out, in, *outlen);
 	    } else {
@@ -840,9 +957,14 @@ void corrupt(corrupt_type_t type, char *in, int inlen, char **out, unsigned *out
 	    break;
 	case REASONABLE_RANDOM:
 	    *outlen = inlen;
-	    *out = (char *) malloc(*outlen);
+	    if(*outlen != 0)
+		*out = (char *) malloc(*outlen);
+	    else
+		*out = malloc(1);
+
 	    for (lup=0;lup<*outlen;lup++)
 		(*out)[lup] = (char) (rand() % 256);
+
 	    break;
 	case REALLYBIG:
 	    *outlen = rand() % 50000;
@@ -895,7 +1017,7 @@ void sendbadsecond(char *mech, void *rock)
 
     if ((hp = gethostbyname(myhostname)) == NULL) {
 	perror("gethostbyname");
-	fatal("");
+	fatal("can't gethostbyname");
     }
 
     addr.sin_family = 0;
@@ -911,14 +1033,14 @@ void sendbadsecond(char *mech, void *rock)
 			0,
 			&clientconn)!= SASL_OK) fatal("sasl_client_new() failure");
 
-    set_properties(clientconn);
+    set_properties(clientconn, NULL);
 
     if (sasl_server_new(service, myhostname, NULL,
 			buf, buf, NULL, 0, 
 			&saslconn) != SASL_OK) {
-	fatal("");
+	fatal("can't sasl_server_new");
     }
-    set_properties(saslconn);
+    set_properties(saslconn, NULL);
 
     do {
 	result = sasl_client_start(clientconn, mech,
@@ -937,7 +1059,7 @@ void sendbadsecond(char *mech, void *rock)
 	fatal("sasl_client_start() error");
     }
 
-    if (mystep == send->step)
+    if (mystep == send->step && outlen)
     {
 	memcpy(buf, out, outlen);
 	corrupt(send->type, buf, outlen, &tmp, &outlen);
@@ -1111,7 +1233,8 @@ void sendbadsecond(char *mech, void *rock)
 
 /* Authenticate two sasl_conn_t's to eachother, validly.
  * used to test the security layer */
-int doauth(char *mech, sasl_conn_t **server_conn, sasl_conn_t **client_conn)
+int doauth(char *mech, sasl_conn_t **server_conn, sasl_conn_t **client_conn,
+           const sasl_security_properties_t *props)
 {
     int result, need_another_client = 0;
     sasl_conn_t *saslconn;
@@ -1136,7 +1259,7 @@ int doauth(char *mech, sasl_conn_t **server_conn, sasl_conn_t **client_conn)
 
     if ((hp = gethostbyname(myhostname)) == NULL) {
 	perror("gethostbyname");
-	fatal("");
+	fatal("can't gethostbyname");
     }
 
     addr.sin_family = 0;
@@ -1153,14 +1276,14 @@ int doauth(char *mech, sasl_conn_t **server_conn, sasl_conn_t **client_conn)
 			&clientconn)!= SASL_OK) fatal("sasl_client_new() failure");
 
     /* Set the security properties */
-    set_properties(clientconn);
+    set_properties(clientconn, props);
 
     if (sasl_server_new(service, myhostname, NULL,
 			buf, buf, NULL, 0, 
 			&saslconn) != SASL_OK) {
-	fatal("");
+	fatal("can't sasl_server_new");
     }
-    set_properties(saslconn);
+    set_properties(saslconn, props);
 
     do {
 	result = sasl_client_start(clientconn, mech,
@@ -1252,6 +1375,16 @@ void cleanup_auth(sasl_conn_t **client, sasl_conn_t **server)
     sasl_done();
 }
 
+
+const sasl_security_properties_t int_only = {
+    0,
+    1,
+    8192,
+    0,
+    NULL,
+    NULL	    
+};
+
 void testseclayer(char *mech, void *rock __attribute__((unused))) 
 {
     sasl_conn_t *sconn, *cconn;
@@ -1260,16 +1393,29 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     const char *txstring = "THIS IS A TEST";
     const char *out, *out2;
     char *tmp;
+    const sasl_security_properties_t *test_props[2] =
+                                          { &security_props, &int_only };
+    const unsigned num_properties = 2;
+    unsigned i;
+    const sasl_ssf_t *this_ssf;
     unsigned outlen, outlen2, totlen;
     
     printf("%s --> security layer start\n", mech);
-    
 
+    for(i=0; i<num_properties; i++) {
+        
     /* Basic crash-tests (none should cause a crash): */
-    if(doauth(mech, &sconn, &cconn) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i]) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
+    if(sasl_getprop(cconn, SASL_SSF, (const void **)&this_ssf) != SASL_OK) {
+	fatal("sasl_getprop in testseclayer");
+    }
+
+    printf("  Testing SSF: %d (requested %d/%d)\n", (unsigned)(*this_ssf),
+	   test_props[i]->min_ssf, test_props[i]->max_ssf);
+
     sasl_encode(NULL, txstring, strlen(txstring), &out, &outlen);
     sasl_encode(cconn, NULL, strlen(txstring), &out, &outlen);
     sasl_encode(cconn, txstring, 0, &out, &outlen);
@@ -1287,7 +1433,7 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     cleanup_auth(&sconn, &cconn);
 
     /* Basic I/O Test */
-    if(doauth(mech, &sconn, &cconn) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i]) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
@@ -1304,7 +1450,7 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     cleanup_auth(&sconn, &cconn);
 
     /* Split one block and reassemble */
-    if(doauth(mech, &sconn, &cconn) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i]) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
@@ -1334,13 +1480,14 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
 
     strcat(buf2, out);
     if(strcmp(buf2, txstring)) {
+	printf("Exptected '%s' but got '%s'\n", txstring, buf2);
 	fatal("did not get correct string back after 2 sasl_decodes");
     }
 
     cleanup_auth(&sconn, &cconn);
 
     /* Combine 2 blocks */
-    if(doauth(mech, &sconn, &cconn) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i]) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
@@ -1377,7 +1524,7 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     cleanup_auth(&sconn, &cconn);
 
     /* Combine 2 blocks with 1 split */
-    if(doauth(mech, &sconn, &cconn) != SASL_OK) {
+    if(doauth(mech, &sconn, &cconn, test_props[i]) != SASL_OK) {
 	fatal("doauth failed in testseclayer");
     }
 
@@ -1427,7 +1574,9 @@ void testseclayer(char *mech, void *rock __attribute__((unused)))
     }
 
     cleanup_auth(&sconn, &cconn);
-
+    
+    } /* for each properties type we want to test */
+     
     printf("%s --> security layer OK\n", mech);
     
 }
@@ -1456,7 +1605,7 @@ void foreach_mechanism(foreach_t *func, void *rock)
 
     if ((hp = gethostbyname(myhostname)) == NULL) {
         perror("gethostbyname");
-        fatal("");
+        fatal("can't gethostbyname");
     }
 
     addr.sin_family = 0;
@@ -1519,11 +1668,12 @@ void test_serverstart()
     struct hostent *hp;
     char buf[8192];
 
-    if (sasl_server_init(emptysasl_cb,"TestSuite")!=SASL_OK) fatal("");
+    if (sasl_server_init(emptysasl_cb,"TestSuite")!=SASL_OK)
+	fatal("can't sasl_server_init in test_serverstart");
 
     if ((hp = gethostbyname(myhostname)) == NULL) {
         perror("gethostbyname");
-        fatal("");
+        fatal("can't gethostbyname in test_serverstart");
     }
 
     addr.sin_family = 0;
@@ -1535,7 +1685,7 @@ void test_serverstart()
     if (sasl_server_new("rcmd", myhostname, NULL,
 			buf, buf, NULL, 0, 
 			&saslconn) != SASL_OK) {
-	fatal("");
+	fatal("can't sasl_server_new in test_serverstart");
     }
 
 
@@ -1616,20 +1766,24 @@ void test_seclayer()
 void create_ids(void)
 {
     sasl_conn_t *saslconn;
-    int i,result;
+    int result;
     struct sockaddr_in addr;
     struct hostent *hp;
     char buf[8192];
+#ifdef DO_SASL_CHECKAPOP
+    int i;
     const char challenge[] = "<1896.697170952@cyrus.andrew.cmu.edu>";
     MD5_CTX ctx;
     unsigned char digest[16];
     char digeststr[32];
+#endif
 
-    if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK) fatal("");
+    if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK)
+	fatal("can't sasl_server_init in create_ids");
 
     if ((hp = gethostbyname(myhostname)) == NULL) {
         perror("gethostbyname");
-        fatal("");
+        fatal("can't gethostbyname in create_ids");
     }
 
     addr.sin_family = 0;
@@ -1641,7 +1795,7 @@ void create_ids(void)
     if (sasl_server_new("rcmd", myhostname, NULL,
 			buf, buf, NULL, 0,
 			&saslconn) != SASL_OK)
-	fatal("");
+	fatal("can't sasl_server_new in create_ids");
     
     /* Try to set password then check it */
 
@@ -1667,6 +1821,7 @@ void create_ids(void)
 	fatal("sasl_user_exists found nonexistant username");
 
     /* Test sasl_checkapop */
+#ifdef DO_SASL_CHECKAPOP
     _sasl_MD5Init(&ctx);
     _sasl_MD5Update(&ctx,challenge,strlen(challenge));
     _sasl_MD5Update(&ctx,password,strlen(password));
@@ -1684,25 +1839,30 @@ void create_ids(void)
     if(result != SASL_OK)
         fatal("Unable to checkapop password we just set");
     /* End checkapop test */
+#else /* Just check that checkapop is really turned off */
+    if(sasl_checkapop(saslconn, NULL, 0, NULL, 0) == SASL_OK)
+	fatal("sasl_checkapop seems to work but was disabled at compile time");
+#endif
 
     /* now delete user and make sure can't find him anymore */
     result = sasl_setpass(saslconn, username, password,
 			  strlen(password), NULL, 0, SASL_SET_DISABLE);
     if (result != SASL_OK)
 	fatal("Error disabling password. Do we have write access to sasldb?");
-    
+
     result = sasl_checkpass(saslconn, username, strlen(username),
 			    password, strlen(password));
     if (result == SASL_OK)
 	fatal("sasl_checkpass got SASL_OK after disableing");
 
+#ifdef DO_SASL_CHECKAPOP
     /* And checkapop... */
     result = sasl_checkapop(saslconn,
                             challenge, strlen(challenge), 
                             buf, strlen(buf));
     if(result == SASL_OK)
         fatal("Checkapop succeeded but should have failed");
-    /* End checkapop */
+#endif
 
     /* try bad params */
     if (sasl_setpass(NULL,username, password, strlen(password), NULL, 0, SASL_SET_CREATE)==SASL_OK)
@@ -1759,12 +1919,13 @@ void test_checkpass(void)
 	fatal("sasl_checkpass() when library not initialized");
     }    
 
-    if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK) fatal("");
+    if (sasl_server_init(goodsasl_cb,"TestSuite")!=SASL_OK)
+	fatal("can't sasl_server_init in test_checkpass");
 
     if (sasl_server_new("rcmd", myhostname,
 			NULL, NULL, NULL, NULL, 0, 
 			&saslconn) != SASL_OK)
-	fatal("");
+	fatal("can't sasl_server_new in test_checkpass");
 
     /* make sure works for general case */
 
@@ -1777,10 +1938,10 @@ void test_checkpass(void)
 		   password, strlen(password)) == SASL_OK)
 	fatal("Suceeded with NULL saslconn");
 
-    /* NULL username */
+    /* NULL username -- should be OK if sasl_checkpass enabled */
     if (sasl_checkpass(saslconn, NULL, strlen(username),
-		   password, strlen(password)) == SASL_OK)
-	fatal("Suceeded with NULL username");
+		   password, strlen(password)) != SASL_OK)
+	fatal("failed check if sasl_checkpass is enabled");
 
     /* NULL password */
     if (sasl_checkpass(saslconn, username, strlen(username),
@@ -1810,6 +1971,7 @@ void usage(void)
            "    g -- gssapi service name to use (default: host)\n" \
 	   "    r -- # of random tests to do (default: 25)\n" \
 	   "    a -- do all corruption tests (and ignores random ones unless -r specified)\n" \
+	   "    n -- skip the initial \"do correctly\" tests\n"
 	   "    h -- show this screen\n" \
            "    s -- random seed to use\n" \
 	   "    M -- detailed memory debugging ON\n" \
@@ -1821,8 +1983,9 @@ int main(int argc, char **argv)
     char c;
     int random_tests = -1;
     int do_all = 0;
+    int skip_do_correct = 0;
     unsigned int seed = time(NULL);
-    while ((c = getopt(argc, argv, "Ms:g:r:h:a")) != EOF)
+    while ((c = getopt(argc, argv, "Ms:g:r:h:an")) != EOF)
 	switch (c) {
 	case 'M':
 	    DETAILED_MEMORY_DEBUGGING = 1;
@@ -1840,6 +2003,9 @@ int main(int argc, char **argv)
 	    random_tests = 0;
 	    do_all = 1;
 	    break;
+	case 'n':
+	    skip_do_correct = 1;
+	    break;
 	case 'h':
 	    usage();
 	    exit(0);
@@ -1856,46 +2022,67 @@ int main(int argc, char **argv)
 
     init(seed);
 
+    printf("Creating id's in sasldb... ");
     create_ids();
-    printf("Created id's in sasldb... ok\n");
     if(mem_stat() != SASL_OK) fatal("memory error");
+    printf("ok\n");
 
+    printf("Checking plaintext passwords... ");
     test_checkpass();
-    printf("Checking plaintext passwords... ok\n");
     if(mem_stat() != SASL_OK) fatal("memory error");
+    printf("ok\n");
 
+    printf("Random number functions... ");
     test_random();
-    printf("Random number functions... ok\n");
     if(mem_stat() != SASL_OK) fatal("memory error");
+    printf("ok\n");
 
+    printf("Testing base64 functions... ");
     test_64();
-    printf("Tested base64 functions... ok\n");
     if(mem_stat() != SASL_OK) fatal("memory error");
+    printf("ok\n");
 
+    printf("Testing auxprop functions... ");
+    test_props();
+    if(mem_stat() != SASL_OK) fatal("memory error");
+    printf("ok\n");
+
+    printf("Tests of sasl_server_init()... ");
     test_init();
-    printf("Tests of sasl_server_init()... ok\n");
     if(mem_stat() != SASL_OK) fatal("memory error");
-
+    printf("ok\n");
+    
+    printf("Testing sasl_listmech()... \n");
     test_listmech();
-    printf("Tests of sasl_listmech()... ok\n");
-    if(mem_stat() != SASL_OK) fatal("memory error");
-
-    test_serverstart();
-    printf("Tests of serverstart... ok\n");
     if(mem_stat() != SASL_OK) fatal("memory error");
+    printf("Testing sasl_listmech()... ok\n");
 
+    if(!skip_do_correct) {
+	printf("Testing serverstart (correct)...\n");
+	test_serverstart();
+	printf("Tests of serverstart (correct)... ok\n");
+	if(mem_stat() != SASL_OK) fatal("memory error");
+    } else {
+	printf("Tests of serverstart (correct)... skipped\n");
+    }
+    
     /* FIXME: do memory tests below here on the things
      * that are MEANT to fail sometime. */
     if(do_all) {	
+	printf("All corruption tests...\n");
 	test_all_corrupt();
 	printf("All corruption tests... ok\n");
     }
     
     if(random_tests) {
+	printf("Random corruption tests...\n");
 	test_rand_corrupt(random_tests);
 	printf("Random tests... ok\n");
+    } else {
+	printf("Random tests... skipped\n");
     }
 
+    printf("Testing security layer...\n");
     test_seclayer();
     printf("Tests of security layer... ok\n");
 
